
# from ..common.common_imports import *
import time
from dataclasses import dataclass
from typing import Callable, List, Tuple, Union
from ..common.colors import *
from evdspy.EVDSlocal.components.options_class import SingletonOptions
from evdspy.EVDSlocal.config.config import *
from evdspy.EVDSlocal.common.files import *
from evdspy.EVDSlocal.setup_project.user_series_config import *
from evdspy.EVDSlocal.setup_project.user_series_config import formulas_dict, frequency_dict
from evdspy.EVDSlocal.setup_project.user_setup_helper import series_title
from evdspy.EVDSlocal.setup_project.user_options_config import items_from_user_config, \
    explanations_config, default_answers_config, check_funcs_options, transform_answers_options
from evdspy.EVDSlocal.config.config import config
# ----------------------------------------------------SetupInputs---------------------------
# -----------------------------------------------------------------------------------------
# SetupInputsConfig = namedtuple('SetupInputsConfig', 'cache_freq gl_date_start gl_date_end')
# SetupInputsSeries = namedtuple('SetupInputsSeries',
#                                'data_folder subject prefix ecodes frequency formulas aggregateType')
cache_freq = SingletonOptions().options_.default_cache
gl_date_start = SingletonOptions().options_.default_start_date
gl_date_end = SingletonOptions().options_.default_end_date
avoid_absolute_paths = SingletonOptions().options_.avoid_absolute_paths
@dataclass
class SetupInputsConfig:
    cache_freq: str = SingletonOptions().options_.default_cache
    gl_date_start: str = SingletonOptions().options_.default_start_date
    gl_date_end: str = SingletonOptions().options_.default_end_date
    avoid_absolute_paths: str = SingletonOptions().options_.avoid_absolute_paths
@dataclass
class SetupInputsSeries:
    data_folder: str
    subject: str
    prefix: str
    ecodes: tuple
    frequency: str
    formulas: str
    aggregateType: str
# -----------------------------------------------------------------------------------------
from enum import Enum
# ----------------------------------------------------configType---------------------------
class configType(Enum):
    Series = 1
    Config = 2
# ----------------------------------------------------InputItem----------------------
@dataclass
class InputItem:
    msg: str
    long_msg: str = ""
    check_func: Callable = TrueFunc
    transform_func: Callable = same
    optional: bool = True
    # -----------------------------------------------------------------------------------------
"""
            GetInputFromUser
"""
# -----------------------------------------------------------------------------------------
from abc import ABC, abstractmethod
# ----------------------------------------------------GetInputFromUser----------------------
@dataclass
class GetInputFromUser(ABC):
    items: List[InputItem]
    config_type: configType
    @abstractmethod
    def ask_user(self, msg, check_func):
        pass
    @abstractmethod
    def get_from_user(self, item, explanation, check_func):
        pass
    @abstractmethod
    def process(self):
        pass
@dataclass
class GetInputFromUserBase():
    items: List[InputItem]
    config_type: configType
    def ask_user(self, msg, check_func):
        ans = input(msg)
        if not check_func(ans):
            return self.ask_user(msg, check_func)
        return ans
    def get_from_user(self, item, explanation, check_func):
        window = "*" * 50 + "\n"
        msg = f"{window}**{item} \n {explanation} :"
        if config.current_mode_is_test:
            return 'pytest running'
        # ans = input(msg)
        ans = self.ask_user(msg, check_func)
        if not ans.strip():
            msg = "==> Continued with default value ..."
            print_with_updating_style(msg)
        print_with_creating_style(f"{window}")
        return str(ans)
    def process(self):
        assert config.current_mode_is_test is False, "config.current_mode_is_test is False"
        if config.current_mode_is_test:
            return get_default_setup_answers(self.config_type)
        user_typed: List = []
        for item in self.items:
            ans = self.get_from_user(item.msg, item.long_msg, item.check_func)
            """ We will transform users answer same / split for now """
            ans = item.transform_func(ans)
            user_typed.append(ans)
        # user_typed = list(map(lambda x: x , user_typed))
        self.user_typed = user_typed
        return tuple(user_typed)
# ----------------------------------------------------GetInputFromUserConfig----------------------
@dataclass
class GetInputFromUserConfig(GetInputFromUserBase):
    """  GetInputFromUserConfig  """
    items: List[InputItem]
    config_type: configType = configType.Config
# ----------------------------------------------------GetInputFromUserSeries----------------------
@dataclass
class GetInputFromUserSeries(GetInputFromUserBase):
    """  GetInputFromUserSeries  """
    items: List[InputItem]
    config_type: configType = configType.Series
# ----------------------------------------------------create_Input_Items----------------------
def create_Input_Items(
        config_type: configType,
        items_from_user: List, explanations: List,
        check_funcs: List[Callable],
        transform_answers: List[Callable],
):
    items: List[InputItem] = []
    for msg, explanation, check_func, transform_func in zip(items_from_user, explanations, check_funcs,
                                                            transform_answers):
        items.append(
            InputItem(
                msg=msg,
                long_msg=explanation,
                check_func=check_func,
                transform_func=transform_func
            )
        )
    GetInputFromUserClassType = {
        configType.Series: GetInputFromUserSeries,
        configType.Config: GetInputFromUserConfig
    }[config_type]
    GIF = GetInputFromUserClassType(
        items=items, config_type=config_type
    )
    return GIF.process()
from typing import List
from evdspy.EVDSlocal.common.folder_name_checks import check_remove_back_slash
# -------------- GOES to Utils  ------------------------
# split_items: any = lambda text : list(
#         text.translate(text.maketrans({x: "-" for x in "[,-/\n;]~"})).split("-"))
def split_items_multi(series_list: Union[str, List, Tuple]) -> str:
    """ 1.1 """
    def split_items(series_list) -> tuple:
        if isinstance(series_list, tuple([list, tuple])):
            return series_list
        return tuple(series_list.translate(series_list.maketrans({x: "-" for x in "[,-/\n;]~"})).split("-"))
    """ 1 """
    if isinstance(series_list, str):
        if not series_list.strip():
            series_list: list = default_answers_series
        else:
            series_list: list = list(split_items(series_list))
    if isinstance(series_list, tuple([list, tuple])):
        series_list_str: str = "\n".join(series_list)
    else:
        series_list_str: str = "\n".join(list(split_items(series_list)))
    return series_list_str
# -------------- GOES to Utils  ------------------------
def content_from_SetupInputsSeries(SI: SetupInputsSeries):
    """ S E R I E S """
    """
    namedtuple('SetupInputsSeries', 'data_folder subject prefix ecodes frequency formulas aggregateType')
    """
    def create_content(folder, subject, prefix, series_list: Union[Tuple, List, str]):
        folder_path = check_remove_back_slash(folder)
        # folder_path_series = str(Path().absolute() / folder / 'series.txt')
        abs_path = str(Path().absolute() / folder_path)
        series_list_str: str = split_items_multi(series_list)
        r = f"""{mainSepBegin}
foldername : {folder_path}
abs_path : {abs_path} # will check again before saving requests from the server it may be replaced by ...WD.../DataSeries/{folder_path}
subject  : {subject}
prefix   : {prefix}
frequency : {SI.frequency} # {frequency_dict[int(SI.frequency)]}
formulas : {SI.formulas} # {formulas_dict[int(SI.formulas)]}
aggregateType : {SI.aggregateType}
------------SERIES CODES------------------
{series_list_str}
------------/SERIES CODES------------------
{mainSepEnd}
{GSEP}"""
        return r
    return create_content(SI.data_folder, SI.subject, SI.prefix, tuple(SI.ecodes))
def content_from_SetupInputsOptions(SI: SetupInputsConfig):
    """ O P T I O N S  """
    """
    SetupInputsConfig = namedtuple('SetupInputsConfig', 'cache_freq gl_date_start gl_date_end')
    """
    def get_default(item):
        # SingletonOptions().get_valid_value("default_cache")
        # SingletonOptions().get_valid_value("default_start_date")
        # SingletonOptions().get_valid_value("default_end_date")
        cache_freq = SingletonOptions().options_.default_cache
        gl_date_start = SingletonOptions().options_.default_start_date
        gl_date_end = SingletonOptions().options_.default_end_date
        avoid_absolute_paths = SingletonOptions().options_.avoid_absolute_paths
        # result = SI.__getattribute__(item)
        # result = SI.__dict__.get(item, None)
        result = getattr(SI, item, None)
        if not result:
            result = locals()[item]
        return result
    def create_content():
        cache_freq = get_default("cache_freq")
        gl_date_start = get_default("gl_date_start")
        gl_date_end = get_default("gl_date_end")
        avoid_absolute_paths = get_default("avoid_absolute_paths")
        """ Options """
        r = f"""#Global Options File
# G L O B A L   O P T I O N S   F I L E   -------------------------------------------------------
cache_freq : {cache_freq}
gl_date_start : {gl_date_start}
gl_date_end : {gl_date_end}
avoid_absolute_paths : {avoid_absolute_paths}
{GSEP}"""
        print(r)
        return r
    return create_content()
def content_from_SetupInputsSeries_MainFile(SI: SetupInputsSeries):
    cont = content_from_SetupInputsSeries(SI)
    content = f"{cont}"
    return content
def content_from_SetupInputsOptions_MainFile(SI: SetupInputsConfig):
    cont = content_from_SetupInputsOptions(SI)
    content = f"{cont}"
    return content
# ----------------------------------------------------create_config_series----------------------
def create_config_series(SI: SetupInputsSeries):
    """ Main Aggregated Configuration file for all series """
    config_series_main_file = 'config_series.cfg'
    SetupInputsSeriesRepr = content_from_SetupInputsSeries_MainFile(SI)
    if not is_file(config_series_main_file):
        SetupInputsSeriesRepr = series_title + "\n" + SetupInputsSeriesRepr
    WriteAdd(config_series_main_file, SetupInputsSeriesRepr)
    """ Individual Series ( Each call create only one set of series) """
    create_folder(SI.data_folder)
    series_file_content = content_from_SetupInputsSeries(SI)
    Write(str(Path() / SI.data_folder / 'series.txt'), series_file_content)
    return True
# ----------------------------------------------------create_config_config----------------------
def create_config_config(SI: SetupInputsConfig):
    """ Main Aggregated Configuration file for all USER OPTIONS """
    config_options_main_file = 'options.cfg'
    SetupInputsSeriesRepr = content_from_SetupInputsOptions_MainFile(SI)
    Write(config_options_main_file, SetupInputsSeriesRepr)
    print_with_creating_style("writing")
    print_with_creating_style(SetupInputsSeriesRepr)
    return True
# ----------------------------------------------------             ----------------------
def get_item_for_config_type(obj, config_type):
    return obj[config_type]
# ---------------------------------------------------- InputType ----------------------
@dataclass
class InputType:
    items_from_user: list
    explanations: list
    check_funcs: list
    transform_answers: Union[List[callable], Tuple[callable]]
    applyFunc: Callable  # not from config
    SetupInputs: any  # ...
# ----------------------------------------------------start_setup----------------------
def start_setup(
        config_type: configType = configType.Series,
        onsetup: bool = False,
):
    # assert config_type == configType.Series, "config_type not configType.Series "
    """ S E R I E S """
    Series_obj = InputType(
        items_from_user_series,
        explanations_series,
        check_funcs_series,
        transform_answers_series,
        create_config_series,
        SetupInputsSeries
    )
    """ O P T I O N S """
    Config_obj = InputType(
        items_from_user_config,
        explanations_config,
        check_funcs_options,
        transform_answers_options,
        create_config_config,
        SetupInputsConfig
    )
    instance = {configType.Config: Config_obj, configType.Series: Series_obj}[config_type]
    if not onsetup:
        CIS = create_Input_Items(config_type, instance.items_from_user, instance.explanations, instance.check_funcs,
                                 instance.transform_answers)
        print_with_creating_style(CIS)
    else:
        CIS = tuple("" for _ in items_from_user_config)
    instance.applyFunc(instance.SetupInputs(*CIS))
# ----------------------------------------------------start_setup_series----------------------
def start_setup_series():
    msg = """
    This process is going to add new series to your configuration file (config_series.cfg)....
    Program will ask couple of questions to complete your series block on your file.
    Pressing `Enter` will accept for default values for the questions.
    You may update all information later by opening file with your favorite file editor.
    --------------------------------------
"""
    print_with_creating_style(msg)
    time.sleep(1)
    start_setup(configType.Series)
    return True
# ----------------------------------------------------start_setup_config----------------------
def start_setup_config(onsetup=False):
    file_name_options = "options.cfg"
    msg = """
        This process is going to create user options file  (options.cfg)....
        You may also update all info by opening file with your favorite file editor.
        --------------------------------------
        Program will ask couple of questions to complete your options on your file.
    """
    if onsetup:
        msg = f"""
        creating `{file_name_options}`
"""
        if file_exists_show_and_return(file_name_options):
            return
    print_with_creating_style(msg)
    time.sleep(1)
    start_setup(configType.Config, onsetup)
    return True
# -----------------------------------------------------------------------------------------
"""
            default_setup
"""
# -----------------------------------------------------------------------------------------
def get_default_setup_answers(config_type: configType = configType.Config):
    default_answers = {
        configType.Series: default_answers_series,
        configType.Config: default_answers_config
    }[config_type]
    return default_answers
def default_setup(config_type: configType = configType.Config):
    print_with_updating_style("Default setup starting....")
    CIS = get_default_setup_answers(config_type)
    create_config_series(SetupInputsSeries(*CIS))
    return True
__all__ = [
    'default_setup',
    'create_config_series',
    'start_setup_series',
    'default_setup',
    'folder_creatable_by_adding',
    'folder_format_check',
    'folder_creatable',
    'get_default_setup_answers'
]


# ................................................................. UserRequestUtils
# ...........................................................................................
from evdspy import setup



from evdspy.EVDSlocal.requests_.real_requests import *
# ...........................................................................................
from typing import Callable , Any 



class UserRequestUtils:
    @staticmethod
    def looks_like_datagroup(string: Any):
        if not isinstance(string, str):
            return False
        return 'bie_' in string

    @staticmethod
    def clean_chars(string: str):
        import re
        if UserRequestUtils().looks_like_datagroup(string):
            string = re.sub('[^0-9a-zA-Z._]+', '', string)
            return string
        string = string.replace("_", ".")
        string = re.sub('[^0-9a-zA-Z.]+', '', string)
        return string


def test_clean_chars(capsys):
    with capsys.disabled():
        assert UserRequestUtils.clean_chars("AAA_BBB?*-()$? ") == "AAA.BBB"
        assert UserRequestUtils.clean_chars("AAA..._BBB?*-()$? ") == "AAA....BBB"
        assert UserRequestUtils.clean_chars("bie_BBB?*-()$? ") == "bie_BBB"


def test_cache_or_raw_fnc(capsys):
    with capsys.disabled():
        setup()
        fnc = lambda x: x ** 2
        f = cache_or_raw_fnc(fnc, True)
        f2 = cache_or_raw_fnc(fnc, False)
        assert f(3) == 9
        assert f2(3) == 9
        assert id(f2) != id(f)


# ..................................................................................
cache_ = MyCache().cache

def create_cache_version(fnc: Callable):
    @cache_
    def fnc_cache(*args, **kw):
        return fnc(*args, **kw)

    return fnc_cache


def cache_or_raw_fnc(fnc, cache=False):
    if not cache:
        return fnc
    return create_cache_version(fnc)



from typing import Union, Any, Optional, Tuple, Dict, Literal
import pandas as pd

from evdspy.EVDSlocal.index_requests.user_requests.Request_config import RequestConfig


def initial_api_process_when_given(api_key: Optional[str] = None) -> None:
    from evdspy.EVDSlocal.config.apikey_class import ApikeyClass

    if api_key is None:
        return
    if ApikeyClass().get_valid_api_key(check=False) is False:
        from evdspy.EVDSlocal.initial.load_commands_cmds_to_load import save_apikey

        save_apikey(api_key)


from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    default_start_date_fnc,
    default_end_date_fnc,
)


def get_series(
    index: Union[str, tuple[Any, ...]],
    start_date: str = default_start_date_fnc(),
    end_date: str = default_end_date_fnc(),
    frequency: Union[
        Literal[
            "monthly", "weekly", "annually", "semimonthly", "semiannually", "business" , None 
        ]
    ] = None,
    formulas: Union[Literal["level", "percentage_change", "difference"], None] = None,
    aggregation: Union[
        Literal["avg", "min", "max", "first", "last", "sum", None], None
    ] = None,
    cache: bool = True ,
    proxy: Optional[str] = None,
    proxies: Optional[dict[str, str]] = None,
    debug: bool = False,
    api_key: Optional[str] = None,
) -> Union[pd.DataFrame, RequestConfig]:
    """
    Retrieves economic data series from the specified API and returns it as a pandas DataFrame.
    Parameters
    ----------
    index : str or tuple of str
        The identifier(s) for the data series to fetch. Can be a single string for one series or a tuple of strings for multiple series.
    start_date : str, optional
        The start date for the data retrieval in 'DD-MM-YYYY' format, by default calls default_start_date_fnc().
    end_date : str, optional
        The end date for the data retrieval in 'DD-MM-YYYY' format, by default calls default_end_date_fnc().
    frequency : str, optional
        The frequency at which data should be retrieved.
        monthly | weekly | annually | semimonthly | semiannually | business
    formulas : str or tuple of str, optional
        The computation methods to apply to the data series
        level | percentage_change | difference | year_to_year_percent_change | year_to_year_differences
    aggregation : str or tuple of str, optional
        The aggregation methods to apply to the data, similar to formulas.
        avg |min | max | first | last | sum
    cache : bool, optional
        If True, uses cached data when available to speed up the data retrieval process, by default False.
    proxy : str, optional
        The URL of the proxy server to use for the requests, by default None.
    proxies : dict, optional
        A dictionary of proxies to use for the request, by default None.
    debug : bool, optional
        If True, runs the function in debug mode, providing additional debug information without making a real API request, by default False.
    api_key : str, optional
        The API key required for accessing the data, by default None.
        When it was given for the first time it will be saved to a file for the subsequent requests.
        alternatively it may be saved by save("APIKEY") function or $ evdspy save [from console]
    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the retrieved data series.
    Raises
    ------
    ValueError
        If an invalid API key is provided or required parameters are missing.
    """

    from evdspy.EVDSlocal.index_requests.user_requests import (
        RequestConfig,
        ProxyManager,
        UrlBuilder,
        DataProcessor,
    )
    from evdspy.EVDSlocal.index_requests.user_requests.Api_requester import ApiRequester

    # ............initial_api_process_when_given...............
    initial_api_process_when_given(api_key)
    # ............RequestConfig................................
    config = RequestConfig(
        index=index,
        start_date=start_date,
        end_date=end_date,
        frequency=frequency,
        formulas=formulas,
        aggregation=aggregation,
        cache=cache,
    )
    # ............ProxyManager................................
    proxy_manager = ProxyManager(proxy=proxy, proxies=proxies)
    # ............UrlBuilder..................................
    url_builder = UrlBuilder(config, url_type=None)
    # ............ApiRequester................................
    api_requester = ApiRequester(url_builder, proxy_manager)
    if debug:
        return api_requester.dry_request()
    # ............DataProcessor................................
    data_processor = DataProcessor(api_requester())
    return data_processor()


def test_get_series2(capsys):
    with capsys.disabled():
        # setup()
        df = get_series("TP.ODEMGZS.BDTTOPLAM", cache=False)
        assert isinstance(df, pd.DataFrame)


def t_stream():
    import streamlit as st

    df = get_series("TP.ODEMGZS.BDTTOPLAM", cache=True)
    st.write(df)


__all__ = ("get_series",)




from dataclasses import dataclass
from rich.console import Console
from rich.table import Table
def split_wisely(line: str, sep: str, num: int):
    parts = line.split(sep)
    return parts[0:num]
@dataclass
class Table_:
    def get_color(self, items):
        for item in items:
            if 'false' in item.lower():
                return 'red'
        return 'green'
    def show(self, content: str, title: str, columns: tuple, skiprow=1):
        table = Table(title=title)
        column_num = len(columns)
        for column in columns:
            table.add_column(column, justify="left", style="cyan", no_wrap=True)
        lines = content.splitlines()
        if not lines:
            return
        for item in lines[skiprow:]:
            parts = split_wisely(item, ":", column_num)
            color_condit = self.get_color(parts)
            table.add_row(*parts, style=color_condit)
        console = Console()
        console.print(table)
import typing as t
from typing import List
@dataclass
class Table2_:
    def get_color(self, items):
        for item in items:
            if 'false' in str(item).lower():
                return 'red'
        return 'green'
    def show(self, list_: List[tuple], title: str, columns: t.Union[tuple, None], skiprow=1):
        table = Table(title=title)
        if columns is None:
            columns = ('key', 'value')
        column_num = len(columns)
        for column in columns:
            table.add_column(column, justify="left", style="cyan", no_wrap=True)
        lines = list_
        if not lines:
            return
        for item in lines:
            # parts = split_wisely(item, ":", column_num)
            parts = item
            color_condit = self.get_color(parts)
            table.add_row(*parts, style=color_condit)
        console = Console()
        console.print(table)
# Table2_.show(list_ = [] , title = '' ,columns = () , skiprow=0  )


from evdspy.EVDSlocal.console.proxy_for_menu import get_proxies_env_helper


def test_get_proxies_env_helper(capsys):
    x1 = (["example:80", None, "example:80", None], {'http': 'example:80'})
    x2 = (["example:80", "example:80", "example:80", None], {'http': 'example:80', 'https': 'example:80'})
    x3 = (["example:80", "example:89", "example:90", "example:91"], {'http': 'example:80', 'https': 'example:89'})
    x4 = ([None, None, "example:90", "example:91"], {'http': 'example:90', 'https': 'example:91'})
    liste = [x1, x2, x3, x4]
    with capsys.disabled():
        for x, expected in liste:
            res = get_proxies_env_helper(*x)
            print(res)
            assert res == expected




from evdspy.EVDSlocal.common.files import *
from dataclasses import dataclass
from datetime import datetime
@dataclass
class Report:
    file_name: str = "stats_report.txt"
    @property
    def date_content(self):
        now = datetime.now()
        current_time = now.strftime("%Y-%m-%d %H:%M:%S")
        return current_time
    def add(self, content):
        content = f"{self.date_content} => {content}"
        WriteAdd(self.file_name, content)
def report_test():
    r = Report(file_name="test_stats_report.txt")
    r.add("content")



site_visitors = r"""
#Series_config_file
E V D S P Y  _  C O N F I G  _  F I L E  ---------------------------------------------
#
# This file will be used by evdspy package (python) in order to help updating
# your series.
# Script will be adding this file when you setup a new project.
# Deleting or modifying its content may require to setup configuration from the beginning
# ----------------------------------------------------------------------------------------
#
#About alternative params
# ----------------------------------------------------------------------------------------
          Frequencies
          -----------------
          Daily: 1
          Business: 2
          Weekly(Friday): 3
          Twicemonthly: 4
          Monthly: 5
          Quarterly: 6
          Semiannual: 7
          Annual: 8
          `Formulas`s
          -----------------
          Level: 0
          Percentage change: 1
          Difference: 2
          Year-to-year Percent Change: 3
          Year-to-year Differences: 4
          Percentage Change Compared to End-of-Previous Year: 5
          Difference Compared to End-of-Previous Year : 6
          Moving Average: 7
          Moving Sum: 8
          Aggregate types
          -----------------
          Average: avg,
          Minimum: min,
          Maximum: max
          Beginning: first,
          End: last,
          Cumulative: sum
#Begin_series
---Series---------------------------------
foldername : visitors/annual
abs_path : visitors/annual # will check again before saving requests from the server it might be replaced by ...WD.../DataSeries/visitors\monthly
subject  : visitors
prefix   : EVPY_
frequency : 8 # annually
formulas : 0 # Level
aggregateType : avg
------------SERIES CODES------------------
TP.ODEMGZS.BDTTOPLAM
TP.ODEMGZS.ABD
TP.ODEMGZS.ARJANTIN
TP.ODEMGZS.BREZILYA
TP.ODEMGZS.KANADA
TP.ODEMGZS.KOLOMBIYA
TP.ODEMGZS.MEKSIKA
TP.ODEMGZS.SILI
------------/SERIES CODES------------------
---/Series---------------------------------
--++--
---Series---------------------------------
foldername : visitors/monthly
abs_path : visitors/monthly # will check again before saving requests from the server it might be replaced by ...WD.../DataSeries/visitors\monthly
subject  : visitors
prefix   : EVPY_
frequency : 5 # Monthly
formulas : 0 # Level
aggregateType : avg
------------SERIES CODES------------------
TP.ODEMGZS.BDTTOPLAM
TP.ODEMGZS.ABD
TP.ODEMGZS.ARJANTIN
TP.ODEMGZS.BREZILYA
TP.ODEMGZS.KANADA
TP.ODEMGZS.KOLOMBIYA
TP.ODEMGZS.MEKSIKA
TP.ODEMGZS.SILI
------------/SERIES CODES------------------
---/Series---------------------------------
--++--
"""
test_series_file_content_for_test = site_visitors
example_series = [site_visitors]
__all__ = [
        'example_series',
]



from evdspy.EVDSlocal.manual_requests.prepare import PrepareUrl
def test_prep(capsys):
    p = PrepareUrl(series=('TP_ODEMGZS_NORVEC-8',), api_key='api_key')
    # with capsys.disabled():
    #     p = PrepareUrl(series=('TP_ODEMGZS_NORVEC-8',), api_key='api_key')
    #     #print(p.get_data())


import logging
import sys
from logging.handlers import TimedRotatingFileHandler
import os
from pathlib import Path
from inspect import getframeinfo, stack
from evdspy.EVDSlocal.initial.start_options import DEGUB_NOTICE

NoteShown = False

FORMATTER = logging.Formatter("%(asctime)s â€” %(name)s â€” %(levelname)s â€” %(message)s  ")
LOG_FILE = "LOG_"
"""
This will stop or activate debugging
"""


class Log:
    def __init__(self, name):
        self.name = name

    def get_console_handler(self):
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(FORMATTER)
        return console_handler

    def get_file_handler(self, logger_name):
        log_file_path = Path() / "logs" / "log_files"  # / "log-{}-{}.log"
        if not os.path.exists(log_file_path):
            # return# TODO
            # os.makedirs(log_file_path)
            ...
        log_file_name = log_file_path / "log-{}-{}.log".format(LOG_FILE, logger_name)
        # print(log_file_name)
        self.log_file_path = log_file_path
        self.log_file_name = log_file_name
        file_handler = TimedRotatingFileHandler(log_file_name, when='midnight', encoding="UTF-8")
        file_handler.setFormatter(FORMATTER)
        return file_handler

    def __repr__(self):
        repr = f"""
---------------------------------------
        LOG
---------------------------------------
this is Log class
debug level set Debug
filename : {self.log_file_name}
----------------------------------------
"""
        return repr

    def get_logger(self, logger_name):
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(self.get_file_handler(logger_name))
        logger.propagate = False
        return logger

    def test(self):
        my_logger = self.get_logger(".{}".format(__name__))
        my_logger.debug("a debug message")
        my_logger2 = self.get_logger(".2-{}".format(__name__))
        my_logger2.debug("a debug message")
        v = "test"
        obj = {"value": v}
        d = my_logger.debug
        d(" ")
        m = my_logger
        l = my_logger
        l.info(" test ")
        my_logger.info(" hello dolly   " + v + str(obj))


def debugInfo(s="", *args, **kw):
    caller = getframeinfo(stack()[2][0])
    msg = "---called from : {} - line number :  {} -- {} ".format(caller.filename, caller.lineno, s)
    return msg


def get_debugger(session=None):
    def show_notice(*args):
        if not DEGUB_NOTICE:
            return
        title = "=" * 50 + " DEBUG " + "=" * 50
        print(f"{title}\nInfo : DEBUG_LOG_CANCEL is True therefore debugging was cancelled.\n "
              "You should modify `initial.start_options.py`  file if you like to activate\n"
              f"debugging and logging behavior. if DEGUB_NOTICE is False this message will not visible  \n{title}")

    def do_nothing(*args):
        pass

    def show_once(*args):
        global NoteShown
        if not NoteShown:
            # show_notice_b = update_wrapper(partial(show_notice, *args), show_notice)
            NoteShown = True
            return show_notice
        return do_nothing

    debugOBJ = None

    # debugOBJ = Log("DEBUG_log_")
    # if session:
    #     debugOBJ.logger = debugOBJ.get_logger("DEBUG_log_{}.-{}".format(__name__, session.hash))
    # else:
    #     debugOBJ.logger = debugOBJ.get_logger("DEBUG_log_{}".format(__name__))
    # # deb = debugOBJ.logger.debug
    # # deb2 = debugOBJ.get_logger("DEBUG_URGENT_{}".format(__name__)).debug

    def deb_multi(*args):
        funcs = []
        d = debugInfo(*args)
        args = args + tuple([d])
        for item in args:
            # deb(item)
            ...

    def deb_multi2(*args):
        funcs = []
        for item in args:
            deb2(item)

    # TODO Cancel Debug in prod  
    # if DEBUG_LOG_CANCEL:
    if True:
        deb_multi = do_nothing
        deb_multi2 = do_nothing
        debug = None
    return deb_multi, deb_multi2, debugOBJ


############## short function
deb, deb2, debug = get_debugger()
if __name__ == "__main__":
    Log("t").test()




from rich import inspect
from .datagroups_initial import data_strategy
from evdspy.EVDSlocal.common.files import Write
import typing as t
import json
from .datagroups import get_datagroups_df
def get_datagroups_data(data_strategy_type="json"):
    # csv = GeneralIndexesDatagroups().get_csv()
    buffer_function: t.Callable = data_strategy[data_strategy_type]  # GeneralIndexesDatagroups().get_json
    content_fn: t.Callable = buffer_function(1)
    content: str | t.Callable = content_fn()
    if callable(content):
        # inspect(content, all=True)
        content = content()
        # raise "content"
    if content:
        inspect(content, all=True)
        if isinstance(content, tuple([list, tuple])):
            # content = "\n".join(content)
            content = json.dumps(content)
        Write("subject_list.txt", content)
    else:
        inspect(content)
    get_datagroups_df(data_strategy_type, content)
    return content


from evdspy.EVDSlocal.common.common_imports import *
from evdspy.EVDSlocal.utils.utils_general import *
from evdspy.EVDSlocal.initial.start_options import *
from ..common.colors import *
from evdspy.EVDSlocal.initial.start_options import default_data_folder_name
import time

DefaultDirs = (default_data_folder_name,)

def get_cache_folder():
    pickle_folder = Path.home() / ".cache" / "evdspy"
    return pickle_folder

def create_cache_folder():
    pickle_folder = get_cache_folder()
    create_directory(str(pickle_folder))

def check_folders_setup_necessary():
    dirs = [Path(".") / x for x in DefaultDirs]
    pickle_folder = get_cache_folder()
    dirs.append(pickle_folder)
    return not all([x.is_dir() for x in dirs])

def check_setup():
    return not check_folders_setup_necessary()

def setup_folders():
    if not check_folders_setup_necessary():
        return
    dirs = [Path() / x for x in DefaultDirs]
    pickle_folder = get_cache_folder()
    dirs.append(pickle_folder)
    NEWLINE = "\n"
    def str_(x: Path):
        return str(x.absolute())
    content = NEWLINE.join(tuple(map(str_, dirs)))
    print_with_success_style(f"""
    Folders will bre created
    ..............................
    {content}
""")
    for item in dirs:
        create_directory(str(item))
        print_with_creating_style(f"Creating directory...{item}" )




#------------------------------------check_remove_back_slash----------------
# back_slash = r"\"" # u"\92"
# back_slash = back_slash.replace('"' , "")
back_slash = "\\"
def check_remove_back_slash(folder_name):
    if str(folder_name).startswith(back_slash):
        folder_name = folder_name[1:]
    return folder_name
reg_relative = '^[a-zA-Z0-9_/\\\\]*$'
reg_absolute_and_relative = '^[a-zA-Z0-9_/\\\\:]*$'
import re
def is_relative(folder):
    result = re.match(reg_relative, folder)
    # if result:
    #     print(f"checked ok {folder}")
    # else:
    #     print(f"not ok {folder}")
    return result



from evdspy.EVDSlocal.components.evds_seri_files import EvdsSeri, EvdsSeriesRequest, EvdsSeriesRequestWrapperFromFile, \
    BucketFromSeriesFile
from evdspy.EVDSlocal.messages.error_classes import *
from evdspy.EVDSlocal.common.files import *
from evdspy.EVDSlocal.series_format.populate_series_helper import *
# from evdspy.EVDSlocal.series_format.series_creator import get_approval
from dataclasses import dataclass 

NEW_LINE = "\n"
GSEP = "--++--"
BeginS = "#Begin_series"
series_begins_note = "---Series - --------------------------------"
from evdspy.EVDSlocal.messages.error_messages import *
from typing import Callable
from evdspy.EVDSlocal.config.config import config
from evdspy.EVDSlocal.series_format.series_examples import test_series_file_content_for_test
from evdspy.EVDSlocal.initial.start_options import *
from evdspy.EVDSlocal.series_format.series_examples import example_series
def get_input(msg):
    if config.current_mode_is_test:
        return "test"
    ans = input(msg)
    return str(ans)
def get_approval(file_name):
    msg = f"{file_name} already exists. If you would like to replace it with a new example please type `replace`: "
    ans = get_input(msg)
    if str(ans).lower().strip() == "replace":
        return True
    return False
def create_series_text_file(file_name: str, content=""):
    """General usage"""
    if Path(file_name).is_file() and "-locked" not in file_name.lower():
        if not get_approval(file_name):
            return False
    print(f"creating...{file_name}")
    result, msg = Write(file_name, content)
    print(msg)
    time.sleep(1)
    return result
import random
def create_series_text_example(file_name=default_series_file_name):
    """ main series config file """
    return create_series_text_file(file_name, random.choice(example_series))
@dataclass
class PopulateSeries:
    name: str = "config_series.cfg"
    separator: str = GSEP
    EvdsSeriesRequestWrapper_Type: Callable = EvdsSeriesRequestWrapperFromFile
    injection: bool = False  # for pytest
    def __post_init__(self):
        self.input_file_name = Path() / f"{self.name}"
        if not self.input_file_name.is_file():
            create_series_text_example()
        self.checks()
    def checks(self):
        items_will_be_checked = (
            BeginS in self.series_file_full_content and self.separator in self.series_file_full_content,
            self.series_file_full_content.count(
                series_begins_note) < self.series_file_full_content.count(
                self.separator)
            ,)
        if not all(items_will_be_checked):
            ...
            msg = f"{self.input_file_name} may be corrupted. You may create a new one from the menu. (create series config file)"
            print(msg )
            raise SeriesFileDoesNotExists(msg)

    def extract_group(self, group: str):
        return [x for x in group.split(NEW_LINE) if x != '']
    @property
    def series_file_full_content(self):
        if config.current_mode_is_test:
            return test_series_file_content_for_test
        content = Read(self.input_file_name)
        return content
    @property
    def series_file_content(self):
        if self.injection:
            return self.injected_content
        self.full_content = content = self.series_file_full_content
        if BeginS in self.full_content:
            content = self.full_content.split(BeginS)[1]
        return content
    def check_file(self):
        x = self.input_file_name.is_file()
        if not x:
            print(SeriesFileDoesNotExists(series_file_not_found_error_msg))
            return x
        return x
    def split_series_file(self, content=None):
        """ main function to split content from series file"""
        if content is None:
            """ For pytest or other test processes """
            content = self.series_file_content
        else:
            self.injection = True
            self.injected_content = content
        groups = content.split(self.separator)
        conts = [self.extract_group(group) for group in groups]
        conts = [cont for cont in conts if len(cont) > 1]
        return [self.create_class(cont) for cont in conts]
    def create_class(self, cont):
        def check(x):
            return all(("#" not in x, "--" not in x, ":" not in x, len(x) > 2))
        folder_name = self.extract_prop(cont, "foldername")
        abs_path = self.extract_prop(cont, "abs_path")
        subject = self.extract_prop(cont, "subject")
        prefix = self.extract_prop(cont, "prefix")
        frequency = self.extract_prop(cont, "frequency")
        formulas = self.extract_prop(cont, "formulas")
        aggregateType = self.extract_prop(cont, "aggregateType")
        bfs = BucketFromSeriesFile(folder_name=folder_name,
                                   subject=subject,
                                   prefix=prefix,
                                   frequency=frequency,
                                   formulas=formulas,
                                   aggregateType=aggregateType,
                                   abs_path=abs_path)
        series_codes = [x for x in cont if check(x)]
        objs = [EvdsSeri(x, bfs=bfs) for x in series_codes]
        # def update(item: EvdsSeri):
        #     item.bfs = bfs
        #
        #     return item
        #
        # objs: List[EvdsSeri] = [update(x) for x in objs]
        return self.export_series(objs, bfs)
    def export_series(self, objs, bfs: BucketFromSeriesFile):
        req = EvdsSeriesRequest(series_list=objs, bfs=bfs)
        assert self.EvdsSeriesRequestWrapper_Type is EvdsSeriesRequestWrapperFromFile, "populate 82"
        return self.EvdsSeriesRequestWrapper_Type(bfs=bfs, EvdsSeriesRequest_=req)
    def clean_series_line(self, line):
        line = line.strip()
        line = line.split("#")
        return line[0]
    def extract_prop(self, lines, prop):
        def get_prop(prop, item):
            value = ""
            r = item.split(":")
            if len(r) > 2:
                # "folder:http:xx"
                value = ":".join(r[1:])
            elif len(r) > 1:
                # "filename:xx"
                value = r[1]
            return value.strip()
        lines = [self.clean_series_line(x) for x in lines if ":" in x]
        lines = [get_prop(prop, x) for x in lines if prop in x]
        lines = [x for x in lines if x is not None]
        if len(lines) == 0:
            lines = ["prop_not_found"]
        return lines[0]
    def process(self):
        reqs = self.split_series_file()
        return reqs
__all__ = [
    'PopulateSeries',
]
# req = PopulateSeries().split_series_file()
# print(req )
def check_pop_read():
    ps = PopulateSeries()
    req = ps.split_series_file(content=pop_series_test_content)
    deb2(req)
    return req



from .index_classes import GeneralIndexesDatagroups
from .data_models import DataModelCSV, DataModelJSON
from ..components.options_class import SingletonOptions
from ..requests_.ev_request import EVRequest
# GID_csv = lambda x: GeneralIndexesDatagroups().get_csv
# GID_json = lambda x: GeneralIndexesDatagroups().get_json
data_strategy = {"csv": GeneralIndexesDatagroups(EVRequest_=EVRequest(options_=SingletonOptions())).get_csv,
                 "json": GeneralIndexesDatagroups().get_json}
data_models_dict = {"csv": DataModelCSV, "json": DataModelJSON}



# from rich.console import Console
# from rich.layout import Layout
#
# console = Console()
# layout = Layout()
#
# # Divide the "screen" in to three parts
# layout.split(
#     Layout(name="header", size=3),
#     Layout(ratio=1, name="main"),
#     Layout(size=10, name="footer"),
# )
# # Divide the "main" layout in to "side" and "body"
# layout["main"].split(
#     Layout(name="side"),
#     Layout(name="body", ratio=2),
#
# )
# # Divide the "side" layout in to two
# layout["side"].split(Layout(), Layout())
#
# # console.print(layout)
#
#
# from rich.live import Live
# from time import sleep
#
# with Live(layout, screen=True):
#     while True:
#         sleep(1)


import shutil
from enum import Enum
from dataclasses import dataclass
from datetime import datetime
import requests
from evdspy.EVDSlocal.utils.utils_general import *
from ..components.api_params import get_enum_with_value
from ..components.options_class import load_options, SingletonOptions
from ..common.url_clean import remove_api_key
from evdspy.EVDSlocal.series_format.stats.save_stats_of_requests import Report
from evdspy.EVDSlocal.config.apikey_class import ApikeyClass
from typing import Union
import os
import pickle

# -----------------------------------------------------------------------------
test_result_file_name = str(get_current_dir() / ".." / "requests_" / "test_reg_result")
arg_sep = "_argSEP_"
p = Path()
# pickle_folder = p.absolute() / 'pickles'
pickle_folder = Path.home() / ".cache" / "evdspy"


# -----------------------------------------------------------------------------
class CacheDecider(Enum):
    day = date.today()
    hour = datetime.now().strftime("%H")
    nocache = datetime.now().strftime("%H:%M:%S")
    default = date.today()


options_ = load_options()


def get_default_cache():
    obj = {
        "daily": CacheDecider.day,
        "hourly": CacheDecider.hour,
        "nocache": CacheDecider.nocache,
        "default": CacheDecider.day,
    }
    return obj.get(options_.default_cache, CacheDecider.day)


def cache_represent(choice: CacheDecider):
    obj = {
        CacheDecider.day: "daily",
        CacheDecider.hour: "hourly",
        CacheDecider.nocache: "nocache",
        CacheDecider.default: "daily",
    }
    return obj[choice]


@dataclass
class MyCache:
    decider: Union[CacheDecider, None] = None  # post init
    # get_default_cache()  # get_default_cache()  # CacheDecider.hour
    report_ = Report("stats_requests.txt")
    msg = "will be running now..."

    def __post_init__(self):
        default_cache = SingletonOptions().get_valid_value("default_cache")
        self.decider = get_enum_with_value(
            key=default_cache, enum_=CacheDecider, default_value=CacheDecider.day
        )

    def update_rules_of_cache(self):
        if ApikeyClass().now_testing_is_key_is_valid:
            msg = "since this looks like a test, Program will be changing cache rulse to nocahce at all..."
            print(msg)

    def report(self, content=None):
        if content is None:
            content = f"""
            Cache Info : Current request cache method caches {cache_represent(self.decider)}
"""
        self.report_.add(content)

    def prepare_report_for_cache(self, func, args):
        self.report("Cache loaded...")
        str_args = (x for x in args if isinstance(x, str))
        str_args = tuple(map(remove_api_key, str_args))
        content = f"<function:{func.__name__}>" ", \n".join(str_args)
        self.report(content)

    def cache(self, func):
        self.update_rules_of_cache()

        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            if self.check_cache(func, *args):
                value = self.load_cache(func, *args)
                self.prepare_report_for_cache(func, args)
            else:
                value = func(*args, **kwargs)
                print("Requesting...", self.msg)
                self.save_cache(func, *args, data=value)
            return value

        return wrapper_decorator

    def sort_safe(self, items):
        items = list(items)
        items = sorted(items)
        return items

    def safe_serialize(self, *items):
        safe_list = []
        for item in items:
            if isinstance(item, str):
                safe_list.append(item)
        safe_list = self.sort_safe(safe_list) + [str(self.decider)]
        parameters = arg_sep.join(safe_list)
        return parameters

    def get_hash_alt1(self, st):
        import hashlib
        import base64

        d = hashlib.md5(bytes(st, encoding="utf-8")).digest()
        d = base64.urlsafe_b64encode(d).decode("ascii")
        return d

    def get_func_hash(self, f, *args):
        parameters = self.safe_serialize(*args)
        f_hash = f"{f.__name__}__{parameters}"
        f_hash = self.get_hash_alt1(f_hash)
        return f_hash

    def check_cache(self, f, *args):
        """added below to provide an exception when we check for valid key"""
        self.update_rules_of_cache()
        f_hash = self.get_func_hash(f, *args)
        f_hash_path = Path() / pickle_folder / f_hash
        return check_pickle(f_hash_path)

    def load_cache(self, f, *args):
        f_hash = self.get_func_hash(f, *args)
        f_hash_path = Path(pickle_folder) / f_hash
        return load_pickle(f_hash_path)

    def save_cache(self, f, *args, data=None):
        if isinstance(data, (bool , type(None) )):
            return

        f_hash = self.get_func_hash(f, *args)
        f_hash_path = Path(pickle_folder) / f_hash
        try:
            save_pickle(f_hash_path, data)
        except:
            print("Cache folder may be deleted.\n setup() in order to create...")


def name_format_pickle(file_name: str):
    return Path(f"{file_name}.pickle")


def check_pickle(file_name: str) -> bool:
    file_name_full = name_format_pickle(file_name)
    return os.path.isfile(file_name_full)


def load_pickle(file_name: str, verbose=False) -> Union[bool, requests.models.Response]:
    file_name_full = name_format_pickle(file_name)
    if not file_name_full.is_file():
        return False
    with open(file_name_full, "rb") as infile:
        test_dict_reconstructed: requests.models.Response = pickle.load(infile)
    msg = f"""
    ---------------------------------
    Recent cache was found for this request phrase. It will be loaded instead of making a new redundant request
    You may change cache options in options file.
    cache loaded...({file_name_full})
    ---------------------------------
"""
    if verbose:
        print(msg)
    else:
        print("<CacheFound>")
    return test_dict_reconstructed


def delete_cache_folder():
    path = Path(pickle_folder)
    if not path.is_dir():
        return True
    try:
        shutil.rmtree(path)
        return True
    except Exception as exc:
        print(exc)
    return False


def save_pickle(file_name, data):
    file_name_full = name_format_pickle(file_name)
    save_pickle_helper(file_name_full, data)


def save_pickle_helper(file_name_path, data):
    with open(file_name_path, "wb") as outfile:
        pickle.dump(data, outfile)


def save_pickle_for_test(data):
    return save_pickle(test_result_file_name, data)


def load_test_pickle():
    return load_pickle(test_result_file_name)


def lru_cache_patched(maxsize=128, typed=False):
    import weakref

    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            self_weak = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(maxsize=maxsize, typed=typed)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)

            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)

        return wrapped_func

    if callable(maxsize) and isinstance(typed, bool):
        func, maxsize = maxsize, 128
        return decorator(func)
    return decorator


__all__ = ["lru_cache_patched", "MyCache", "save_pickle_for_test", "load_test_pickle"]



from pathlib import Path
import pandas as pd
from evdspy.EVDSlocal.index_requests.get_series_indexes import (
    default_start_date_fnc,
    default_end_date_fnc,
    get_series,
)
from evdspy.EVDSlocal.utils.utils_general import get_env_api_key
from evdspy.EVDSlocal.utils.utils_test import (
    ApiClassWhileTesting,
    get_api_key_file,
    skip_if_gthub,
    skip_if_not_keyvalid,
    is_df,
)
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    Frequency,
    freq_enum,
    Formulas,
    AggregationType,
    correct_types,
)
from evdspy.EVDSlocal.utils.utils_test import get_api_key_while_testing

# from evdspy.EVDSlocal.index_requests.user_requests import
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    default_start_date_fnc,
    default_end_date_fnc,
)
from evdspy.EVDSlocal.index_requests.user_requests import RequestConfig

import random 

def get_options():
    aggr = [
        "monthly",
        "weekly",
        "annually",
        "semimonthly",
        "semiannually",
        "business",
        None,
    ]
    formulas = ["level", "percentage_change", "difference", None]
    freq = [
        "monthly",
        "weekly",
        "annually",
        "semimonthly",
        "semiannually",
        "business",
        None,
    ]
    def get_one(chs):
        return random.choices(chs, k=1)[0]

    yield {
        "aggregation": get_one(aggr),
        "formulas": get_one(formulas),
        "frequency": get_one(freq),
    }


            
def test_get_series_2(capsys):
    with capsys.disabled(): 
        for _ in range(2) :
            kw = get_options()
            kwargs = tuple(kw)[0]
            print(kwargs)
            try : 
                df = get_series("TP.TIG01", start_date='01-01-2023' ,  debug=False, **kwargs)
                assert isinstance(df, pd.DataFrame)
                
                print(df , type(df ))
            except:
                pass 
            



class BaseException_Request(Exception):
    """BaseException_Request"""
class Req_NotFound(BaseException_Request):
    """Req_NotFound"""
class Req_Unauthorized(BaseException_Request):
    """Req_NotFound"""
class Req_Forbidden(BaseException_Request):
    """Req_NotFound"""
class Req_BadRequest(BaseException_Request):
    """Req_NotFound"""
class Req_TooMany(BaseException_Request):
    """Req_TooMany"""
class Internal(BaseException_Request):
    """Internal"""
class Bad_Gateway(BaseException_Request):
    """Bad_Gateway"""
class ServiceUnavailable(BaseException_Request):
    """ServiceUnavailable"""
class GatewayNotImplemented(BaseException_Request):
    """GatewayNotImplemented"""
REQUEST_ERROR_CODES = {
        400: Req_BadRequest,
        401: Req_Unauthorized,
        403: Req_Forbidden,
        404: Req_NotFound,
        429: Req_TooMany,
        500: Internal,
        502: Bad_Gateway,
        503: ServiceUnavailable,
        504: GatewayNotImplemented
}



from dataclasses import dataclass
from rich import pretty
@dataclass
class CurrentState:
    """ Global State to check loaded functions
    and other global variables  """
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(CurrentState, cls).__new__(cls)
            cls.set_state_on_load(cls)
            cls.set_result_dict(cls)
            cls.initial_sets(cls)
        return cls.instance
    def initial_sets(cls):
        cls.menu_will_open = False
    def set_state_on_load(cls, attr="loaded_menu_on_load", value=True):
        cls.loaded_menu_on_load = False
    def set_state(cls, attr="loaded_menu_on_load", value=True):
        cls.__setattr__(attr, value)
        # self.loaded_menu_on_load = False
    def set_result_dict(cls, attr=None, value=None):
        if attr is None:
            result = {}
            cls.result = result
        else:
            cls.result.update({attr: value})
    def get_result_dict(cls, id=None) -> [dict, any]:
        if id is None:
            result = cls.result
        else:
            result = cls.result.get(id, None)
        return result
    def __repr__(cls):
        return pretty.Pretty(cls.result)
        # return f"CurrentState:{cls.result}"



""" S T A R T   O P T I O N S """
#
#
#   see SingletonOptions.py
#
from ..common.common_imports import *
from pathlib import Path
import sys
import os 

def check_if_this_is_pytest():
    def check():
        path_ = Path(sys.argv[0])
        return "pytest" in str(path_.stem)
    if len(sys.argv) > 0 and check():
        return True
    else:
        return False
#
#
#
if True:
    """temporarily cancel mock requests and do real requests"""
    temp_cancel_mock_request = True
    """series file name """
    default_series_file_name = "config_series.cfg"
    """series data folder name """
    default_data_folder_name = "SeriesData"
    """ for security as a default value it is True to avoid polluting existing folders with unwanted excel files """
    Avoid_Absolute_Paths_ = True
    """ for security we define a prefix to name created excel files to diminish risk of removing existing excel files """
    Default_Prefix_ = "EVPY_"
    # options_folder_name = r"IO"
    """ debug option """
    DEBUG_LOG_CANCEL = False
    """ debug notice """
    DEGUB_NOTICE = True
    """verbose debug """
    DEBUG_PRINT = False  # TODO less verbose debug plus conditional
    """ api key tested hashes folder name  """
    default_arch_folder_name = "APIKEY_FOLDER"
    """ api key tested hashes file name  """
    default_arch_file_name = "arch.txt"
    """default_cache"""
    default_cache = "daily"  # nocache / hourly / monthly
    """default_end_date"""
    default_end_date = "01-12-2030"
    """default_start_date"""
    default_start_date = "01-01-2019"
    # USERNAME = os.getenv("USERNAME")  # or hard coded "Username"
#
#
#
#
#
#
#
#
#
#
""" / S T A R T   O P T I O N S """
# pytest will set this True
current_mode_is_test = check_if_this_is_pytest()
""" GLOBALS """



from dataclasses import field
from enum import Enum
from evdspy.EVDSlocal.components.url_class import URLClass
class RequestType(Enum):
    Request = "request"
    Cache = "cache"
class RequestOrCacheResultInfo():
    safe_url: str = ""
    request_type: RequestType
    columns: tuple
    excel_saved: bool
    url_instance: URLClass
    def __init__(self, safe_url='', request_type=RequestType.Request, columns=(), excel_saved=False, url_instance=None):
        self.safe_url = safe_url
        if request_type is None:
            request_type = RequestType.Request
        if url_instance is None:
            url_instance: URLClass = field(default_factory=URLClass)  # URLClass([])
        self.request_type = request_type
        self.columns = columns
        self.excel_saved = excel_saved
    def get_data(self):
        content = f"""
url : {self.safe_url}
request or cache : {self.request_type.value}
columns : {self.columns}
excel file was saved ? : {self.excel_saved}
"""
        return content
# @dataclass
# class RequestOrCacheResultInfo():
#     safe_url: str = ""
#     request_type: RequestType = field(default=RequestType.Request)
#     columns: field(default_factory=tuple) = field(default_factory=tuple)
#     excel_saved: bool = False
#     url_instance: URLClass = URLClass(())
#
#     def get_data(self):
#         content = f"""
# url : {self.safe_url}
# request or cache : {self.request_type.value}
# columns : {self.columns}
# excel file was saved ? : {self.excel_saved}
# """
#         return content



from .cmd_formats import *
from ..common.common_imports import *
def display_help_messages_md():
    def ReadLocal(filename: str):
        long_des = ""
        with open(filename, "r") as file_:
            long_des = file_.read()
        return long_des
    # return
    from rich.console import Console
    from rich.markdown import Markdown
    print(Path().parent)
    content_md = ReadLocal(str(Path() / Path().parent / "README.md"))
    console = Console()
    # print(content_md)
    md = Markdown(content_md)
    console.print(md)
    return content_md
    # content_md = str(content_md).encode('utf8')
    # md = Markdown(str(content_md))
    # console.print(md)
def display_help_messages():
    from .current_menu_message import current_menu_appears
    """ Some help to show some useful commands of evdspy package """
    ##############  commands #################
    msg = rf"""
--------importing -----
from evdspy  import *
-------------
menu() :
-------------
            Quickly setup your work environment run commands from the menu. \
            Currently you may choose one below or create your own functions and \
            add to this list following instructions on Readme.md
${current_menu_appears}
help_evds() :
-------------
    to see a list of initial commands
    to check setup and reinitialize some installations.
check():
-------------
    see current installation status
setup()   :
-------------
    # this function helps you setup your project by asking your preferences step by step
    setup()
get():
-------------
    # this will check for your current config_series.cfg file
    # if proper data series codes are given it will either download them
    # or use latest cache from your local environment
    # to provide other cache options such as nocache / daily / hourly you may change your
    # defaults or give arguments such as
        get(cache="daily")
        get(cache = "nocache")
        or to go with default from your option file
        get()
save("MyApiKey" ):
-------------
    # saves your api key to use while requesting data from the server.\
    # creates an APIKEY_FOLDER and saves some hash function of your api key for security.
    # If you are on a public computer we suggest you delete this folder when you are done requesting data.
    save("MyApiKey")
save():
-------------
    When you call it with not argument program will ask for your key and do the same
    above
create_series_file():
-------------
    # creates example `config_series.cfg` evdspy input file (EIF) formatted
    # you may modify it accordingly, you may user separator of `--++--` between different series groups
    create_series_file()
    # or
    csf()
menu() :
-------------
Quickly setup your work environment run commands from the menu.
Currently you may choose one below or create your own functions and
add to this list following instructions on Readme.md
${current_menu_appears}
-----------------------------------------------------------------------
# from evdspy import *         #(if you have not imported evdspy yet )
menu()                         # to see the menu()
-----------------------------------------------------------------------
"""
    ############## END of commands #################
    ################ p r i n t ###############
    print(msg)
    ################            #######
from rich import print
from ..common.colors import *
def welcome_message():
    msg = """
    Welcome to evsdpy python interface...
    --------------------------------------------------------------------------------------------------
    This message appears when you use evdspy as a command line prompt.
    This welcome and help message will show up when you use `evdspy` command with no parameters...
    --------------------------------------------------------------------------------------------------
    Some parameters are listed below. If there is another command you are looking for you may check the menu option.
    $ evdspy setup
    --------------
        creates initial folders for your environment to save data and caches
    $ evdspy menu
    --------------
        Loads the menu
    $ evdspy create series
    --------------
        Creates series file (leaves untouched if exists)
    $ evdspy create options
    --------------
        creates options on the current folder
    $ evdspy save
    --------------
        asks for your api key to save a file in your environment named `APIKEY_FOLDER`
    $ evdspy get
    --------------
        makes request from EVDS API and creates excel files regarding information on your series file
    $ evdspy help
    --------------
        shows help and some documentation from command line
"""
    print_with_success_style(msg)


from enum import Enum
from typing import Optional, Union
import os

import traceback

from evdspy.EVDSlocal.config.dotenv import load_env
#
# try :
#
#     from dotenv import load_dotenv
#     load_dotenv()
# except Exception:
#
#     print("""[solution]
#           $ pip install python-dotenv
#           """)
#     traceback.print_exc()
#     raise ModuleNotFoundError()


"""
    ========================== load_env ==========================
"""
load_env()


def check_env(key_name: str, default_=None):
    value = os.getenv(key_name)
    if value:
        return value
    return default_


def default_start_date_fnc():
    return check_env("START_DATE", "01-01-2000")


def default_end_date_fnc():
    return check_env("END_DATE", "01-01-2100")


class AggregationType(Enum):
    """
    avg : "avg"
    min : "min"
    max : "max"
    first : "first"
    last : "last"
    sum : "sum"
    """

    avg = "avg"
    min = "min"
    max = "max"
    first = "first"
    last = "last"
    sum = "sum"

    def __str__(self):
        return self.value


class Formulas(Enum):
    """Formulas
    Level: 0
    Percentage change: 1
    Difference: 2
    Year-to-year Percent Change: 3
    Year-to-year Differences: 4
    Percentage Change Compared to End-of-Previous Year: 5
    Difference Compared to End-of-Previous Year : 6
    Moving Average: 7
    Moving Sum: 8
    """

    level = 0
    percentage_change = 1
    difference = 2
    year_to_year_percent_change = 3
    year_to_year_differences = 4
    percentage_change_compared = 5
    difference_compared = 6
    moving_average = 7
    moving_sum = 8

    def __str__(self):
        return self.value

    @classmethod
    def from_str(cls, string: str):
        """Converts a string to a corresponding enum member."""
        mapping = {
            "level": cls.level,
            "percentage change": cls.percentage_change,
            "difference": cls.difference,
            "year to year percent change": cls.year_to_year_percent_change,
            "year to year differences": cls.year_to_year_differences,
            "percentage change compared": cls.percentage_change_compared,
            "difference compared": cls.difference_compared,
            "moving average": cls.moving_average,
            "moving sum": cls.moving_sum,
        }
        normalized_string = string.lower().replace("_", " ").strip()
        if normalized_string in mapping:
            return mapping[normalized_string]
        else:
            raise ValueError(f"Unknown formula type: {string}")


class Frequency(Enum):
    daily = 1
    business = 2
    weekly = 3  # Friday
    semimonthly = 4
    monthly = 5
    quarterly = 6
    semiannually = 7
    annual = 8
    annually = 8

    def __str__(self):
        return self.value

    def __call__(self, *args, **kwargs):
        return f"&frequency={self.value}"


def freq_enum(frequency: Union[str, int]) -> str:
    def get_enum(value: str):
        obj = {
            "daily": Frequency.daily,
            "business": Frequency.business,
            "weekly": Frequency.weekly,
            "semimonthly": Frequency.semimonthly,
            "monthly": Frequency.monthly,
            "quarterly": Frequency.quarterly,
            "semiannually": Frequency.semiannually,
            "annual": Frequency.annually,
            "annually": Frequency.annually,
        }
        return obj.get(str(value).lower(), Frequency.daily)

    if isinstance(frequency, int):
        return f"&frequency={frequency}"
    return get_enum(frequency)()


def correct_types(value: Optional[tuple], enum_class) -> Union[str, int, tuple]:
    if value is None:
        return value
    # first_type = type(next(iter(enum_class)).value)
    if isinstance(value, tuple):
        return tuple(map(lambda x: correct_types(x, enum_class), value))
    if isinstance(value, enum_class):
        return value.value
    if isinstance(value, str) and hasattr(enum_class, "from_str"):
        return enum_class.from_str(value).value
    else:
        return value




# ------------------------------------------------------------------------------
import requests as requests
from evdspy.EVDSlocal.utils.utils_general import *
from evdspy.EVDSlocal.requests_.my_cache import MyCache
from evdspy.EVDSlocal.messages.error_classes import *
from evdspy.EVDSlocal.config.apikey_class import *
from evdspy.EVDSlocal.config.config import config
from evdspy.EVDSlocal.requests_.my_cache import load_pickle
from evdspy.EVDSlocal.stats.save_stats_of_requests import *
# ------------------------------------------------------------------------------
m_cache = MyCache()
CANCEL_REQUEST_TEMP = config.cancel_request_temp
def mock_request(url, proxies=None) -> requests.models.Response:
    """ pytest will get this mock response object"""
    assert isinstance(url, str)
    file_name = str(Path(__file__).parent / "test_reg_result")
    assert Path(file_name + ".pickle").is_file(), f"file does not exist {file_name}"
    return load_pickle(file_name)



# from ..common.common_imports import *
from ..common.files import *
from dataclasses import dataclass , field 
import os
from pathlib import Path
from evdspy.EVDSlocal.config.config import *
from ..components.options_class import SingletonOptions
from evdspy.EVDSlocal.initial.start_options import default_data_folder_name
from evdspy.EVDSlocal.common.folder_name_checks import check_remove_back_slash
from evdspy.EVDSlocal.components.api_params import FrequencyEnum, AggregationEnum, get_enum_with_value, FormulasEnum 




# ------------------------------------------------------------------------------
# /*
#               BucketFromSeriesFile
#
# ------------------------------------------------------------------------------
# */
from evdspy.EVDSlocal.components.request_or_cache import RequestOrCacheResultInfo
@dataclass
class BucketFromSeriesFile:
    folder_name: any
    subject: str
    frequency: str
    formulas: str
    aggregateType: str
    prefix: str = Default_Prefix_
    NullType: bool = False
    abs_path: any = None
    xls_name_path: str = "test"
    request_or_cache_resultInfo_: field(default_factory=RequestOrCacheResultInfo) = False
    def __post_init__(self):
        """ Check 2 => after writing avoid \foldername\foldername instead of foldername\foldername
            we previously checked while getting input from user but
            we are checking if user added any absolute path by mistake
        """
        def get_defaults(item, default_value):
            if item == "":
                return default_value
            return item
        # self.frequency = get_defaults(str(self.frequency), default_freq)
        # self.formulas = get_defaults(self.formulas, default_formulas_)
        # self.aggregateType = get_defaults(self.aggregateType, default_AggregateType_)
        from evdspy.EVDSlocal.setup_project.user_series_config import default_freq, default_formulas_, \
            default_AggregateType_
        self.frequency: str = get_defaults(str(self.frequency), default_freq)
        self.frequency = get_enum_with_value(self.frequency, FrequencyEnum, FrequencyEnum.default)
        self.formulas: str = get_defaults(self.formulas, default_formulas_)
        self.formulas: str = get_enum_with_value(self.formulas, FormulasEnum, FormulasEnum.default)
        self.aggregateType = get_defaults(self.aggregateType, default_AggregateType_)
        self.aggregateType: str = get_enum_with_value(self.aggregateType, AggregationEnum, AggregationEnum.default)
        self.folder_name = check_remove_back_slash(self.folder_name)
        self.abs_path = self.get_folder_name(self.folder_name)
        self.xls_name_path = str(Path(self.abs_path) / (self.prefix + '_' + self.subject))
    def __str__(self):
        s = f"""
frequency : {self.frequency}
formulas : {self.formulas}
aggregateType : {self.aggregateType}
folder_name : {self.folder_name}
abs_path : {self.abs_path}
xls_name_path : {self.xls_name_path}
"""
        return s
    def get_folder_name(self, folder_name: str):
        def adapt_folder_name(relative_folder, reduce=False):
            if reduce:
                relative_folder = Path(folder_name).stem
            abs_modified = Path() / default_data_folder_name / relative_folder
            return abs_modified
        if folder_name is None:
            return ""
        if Avoid_Absolute_Paths_ and SingletonOptions().options_.avoid_absolute_paths_user:
            """ Namespace Protection is on """
            r"""
            This is extra protection to prevent mismatching of unwanted paths and populating some files replacing
            old files, if the user is sure absolute paths are correct they can modify
                Avoid_Absolute_Paths_ variable in start_options.py file
                avoids : C:Users\Userx\XdataFolderName
                changes paths to correct form : ...WorkingDirectory..\SeriesData\XdataFolderName
            """
            reduce = False
            if Path(folder_name).is_absolute():
                reduce = True
            return adapt_folder_name(folder_name, reduce=reduce)
        """ Namespace Protection turned off by user """
        print("Avoid_Absolute_Paths_ was cancelled")
        if self.create_folder():
            """ now we will return absolute path if we can successfully create the folder"""
            return self.abs_path
        return adapt_folder_name(self.abs_path, reduce=True)
    def create_folder(self):
        if is_dir(self.abs_path):
            return True
        try:
            os.makedirs(self.abs_path)
            return True
        except:
            if not config.current_mode_is_test:
                print(f"could not create folder {self.abs_path} ")
                # raise BucketFromSeriesFolderCreateError
            return False
from evdspy.EVDSlocal.setup_project.user_series_config import default_freq, default_formulas_, \
    default_AggregateType_
null_BucketFromSeriesFile = BucketFromSeriesFile(folder_name=None,
                                                 subject="test",
                                                 frequency=default_freq,
                                                 formulas=default_formulas_,
                                                 aggregateType=default_AggregateType_,
                                                 prefix="EV_",
                                                 NullType=True)



from ..messages.error_messages import *
from .load_commands_cmds_to_load import *
# ---------------------------------------------------------------/ MAIN LOAD
__all__ = [
    'easy_setup',
    'setup',
    'setup_series',
    "setup_series_steps",
    'help_evds',
    'check',
    'get',
    'h',
    'help_evdspy',
    'help_',
    'create_series_file',
    'csf',
    'show',
    'get_df_test',
    'console',
    'menu',
    'menu_onload',
    'version',
    'save_apikey',
    'save',
    'remove_cache',
    'create_options_file',
    'cof',
    'console_main',
    'get_df_datagroup',
    'get_datagroups_with_code',
    'get_category_name' ,
    'get_series_list_of_subject'
]



# from evdspy.EVDSlocal.tests.test_evds_classes import *
# from evdspy.EVDSlocal.tests.test_evds_classes import *
from evdspy.EVDSlocal.tests.test_main_load import *
# from evdspy.EVDSlocal.tests.test_project_setup import *
from evdspy.EVDSlocal.tests.test_log_session import *
from evdspy.EVDSlocal.tests.test_requests import *
from evdspy.EVDSlocal.tests.test_apikey_checks import *
from evdspy.EVDSlocal.tests.test_categories import *
# from evdspy.EVDSlocal.tests.test_main_cmd import *
# from evdspy.EVDSlocal.tests.test_datagroups import *
#-------------- not necessary for now
#
# from .tests.test_series_files import *
# from .tests.test_manual_series import *



from evdspy.EVDSlocal.common.colors import print_with_info_style
from evdspy.EVDSlocal.config.config import config
width = 60
def make_indent(element, num=width):
    indent = " " * (num - len(element))
    return f"{indent}{element}"
def make_center(element, num=width):
    left = " " * round((num - len(element)) / 2)
    return f"{left}{element}"
def menu_header(self):
    menu_title = make_center("M E N U")
    reminder = "to get the latest version:   pip install -U evdspy"
    reminder = make_indent(reminder)
    version_line = make_indent(config.version)
    logo = make_indent("evdspy @2022")
    line_1 = "-" * width
    print_with_info_style(reminder)
    header = f"""
{line_1}
{version_line}
{line_1}
{logo}
{menu_title}
{line_1}
"""
    return header
def version_logo():
    # menu_title = make_center("M E N U")
    reminder = "for the latest version: pip install -U evdspy"
    reminder = make_indent(reminder)
    version_line = make_indent(config.version, 15)
    logo = make_indent("evdspy @2022", 30)
    line_1 = "-" * width
    logo_version = f"""
    {line_1}
  {logo}  {" " * 5} {version_line}
    {line_1}
    """
    logo_version2 = f"""
    {line_1}
    {version_line}
    {line_1}
    {logo}
    """
    return logo_version, reminder


import random
from dataclasses import dataclass, field
from pathlib import Path
import pandas as pd
from evdspy.EVDSlocal.utils.utils_general import get_random_hash


@dataclass
class FileItem:
    """Class for general use case"""
    dir_path: str
    file_name: str
    full_name: Path = None
    column_names: tuple = None
    summary: str = None
    sample_data: list = None
    df: pd.DataFrame = None
    df_json: str = ""
    short_name: str = field(init=False)
    encoded_name: str = "aa"
    sa_out_file: str = None
    sa_f_out_file: str = None
    cal_out_file: str = None
    sa_df = None
    sa_f_df = None
    cal_df = None
    created_items: list = field(default_factory=list)

    def __post_init__(self):
        self.full_name = Path() / self.dir_path / self.file_name
        self.created_items.append(self)

    def __repr__(self):
        return f"""
FileItem(
    dir : {self.dir_path} , 
    file_name : {self.file_name}
    full_name : {self.full_name}
    sa_out_file : {self.sa_out_file}
    calendar_out_file : {self.cal_out_file}
)

{self.column_names}

"""


def make_eng(text: str) -> str:
    # return text
    dict_ = {
            'ş': 's',
            'ç': 'c',
            'ı': 'i',
            'ü': 'u',
            'ö': 'o',
            'ğ': 'g',
            'İ': 'I',
            'Ş': 'S',
            'Ğ': 'G',
            'Ç': 'C',
            'Ü': 'U',
            'Ö': 'O',

    }

    return text.translate(text.maketrans(dict_))


def file_items_update(files):
    def update(file_item: FileItem):
        file_item.short_name = make_eng(str(file_item.full_name.stem))
        file_item.encoded_name = file_item.short_name
        return file_item

    return tuple(map(update, files))


def mock_file_items():
    items = []
    random.seed(954)
    exts = ("xlsx", "mat", "m", "xml")
    folders = (".", "A", "B", "C")
    for item in range(10):
        name = get_random_hash(5)
        f = FileItem(random.choice(folders), name + "." + random.choice(exts))
        items.append(f)
    return items




#### 1 ###############
import base64
import functools
# from babel.numbers import format_decimal
import os.path
from pathlib import Path
import sys
indent = " " * 15
ENCODING = "utf-8"
#### 2 ###############
import os
#### GLOBALS ###############
#### -- ###############
from datetime import date
today = date.today()
stoday = str(today)
import time
import os 

def ls():
    print(os.popen("ls -l").read())

def decode(base64_bytes) -> str:
    if isinstance(base64_bytes, str):
        base64_bytes = str.encode(base64_bytes)
    str_bytes = base64.b64decode(base64_bytes)
    decoded = str_bytes.decode("ascii")
    return decoded
def encode(text: str) -> bytes:
    t_bytes = text.encode("ascii")
    encoded = base64.b64encode(t_bytes)
    return encoded

def replace_recursive(content: str, char: str, new_char: str):
    if char not in content:
        return content
    content = content.replace(char, new_char)
    return replace_recursive(content, char, new_char)
    
def api_key_looks_valid(key: str):
    return isinstance(key, str) and len(key) == 10
class ApiKeyErrorEnvir(BaseException):
    """ApiKeyErrorEnvir"""
def get_env_api_key(check=False):
    key = os.getenv("EVDS_API_KEY")
    if check and not api_key_looks_valid(key):
        from evdspy.EVDSlocal.utils.utils_test import get_api_key_while_testing
        api_key = get_api_key_while_testing()
        if api_key_looks_valid(api_key):
            return api_key
        raise ApiKeyErrorEnvir("ApiKeyErrorEnvir")
    return key
def do_if_callable(f):
    if callable(f):
        f()
def URL_temizle(url: str):
    import string
    return url.translate({ord(c): None for c in string.whitespace})
def arg_acc(argv=None):
    if argv is None:
        argv = sys.argv
    obj = {}
    for index, key in enumerate(argv):
        key = str(key)
        if key.startswith("--"):
            value = None
            if len(argv) > index + 1:
                value = argv[index + 1]
            key = key[2:]
            obj.update({key: value})
    return obj
def get_current_dir():
    return Path(__file__).parent
def get_current_dir2(pat):
    d = pat.absolute()
    return d
def delay(s: int, f1=None, f2=None):
    do_if_callable(f1)
    time.sleep(s)
    do_if_callable(f2)
def ro(amount: float):
    return round(amount)
def bound(f, *args):
    return functools.partial(f, *args)
def create_directory(adres: str) -> None:
    import os
    if not os.path.isdir(adres):
        os.makedirs(adres)
def get_api_key_from_file(file_name):
    lines = False
    file_name_path = Path(file_name)
    if not file_name_path.is_file():
        return False
    with open(file_name, encoding=ENCODING) as file:
        lines = file.readlines()
    if lines is False:
        return False
    return str(lines[0]).strip()  # .split(":")[1]
def global_var_api_key():
    import os
    return "This version does not read global EVDS"
    API_KEY_NAME = "EVDS_API_KEY"
    return os.getenv(API_KEY_NAME)
def get_proxy_from_file(file_name):
    lines = False
    if not file_name:
        return False
    file_name_path = Path(file_name)
    if not file_name_path.is_file():
        return False
    with open(file_name_path, encoding=ENCODING) as file:
        lines = file.readlines()
    if lines is False:
        return False
    return lines[0]  # ":".join(lines[0].split(":")[1:])
def reducer(liste):
    functools.reduce(lambda a, b: a + b, liste)
def reducer_prop(liste, prop="guncel_tutar"):
    # print(liste ,prop , "---" )
    return functools.reduce(lambda a, b: a + b,
                            [getattr(item, prop) for item in liste])
def f2(amount):
    return '{:7,.2f}'.format(amount)
def f0(amount):
    return '{:7,.0f}'.format(amount)
def get_random_hash(num=5):
    import random
    import string
    letters = string.ascii_letters
    randomF = ''.join(random.choice(letters) for i in range(num))
    return randomF




import time
import pandas as pd
from ..components.evds_files import DfColumnsDoesNotMatch
from ..config.config import config
from ..common.colors import print_with_failure_style, print_with_info_style
from .data_models import *
from dataclasses import dataclass
@dataclass
class DFOperations():
    data_model: DataModel
    df: pd.DataFrame = None
    @staticmethod
    def extract_values_from_line(line: str, sep=","):
        return line.split(sep)
    @staticmethod
    def line_gen(buffer: str):
        if not buffer:
            return False
        for item in buffer.split("\n"):
            yield item
    def get_columns_from_first_row(self):
        first_row = self.data_model.data.split("\n")[0].split(",")
        return first_row
    def get_columns(self, df):
        current_columns = df.columns  # rangeIndex
        if isinstance(current_columns, pd.RangeIndex):
            current_columns_len = len([x for x in current_columns])
        else:
            current_columns_len = len(current_columns)
        potential_columns = self.get_columns_from_first_row()
        if len(potential_columns) != current_columns_len:
            raise DfColumnsDoesNotMatch()
        df.columns = self.get_columns_from_first_row()
        UnixTime = 'UNIXTIME'
        if UnixTime in df.columns:
            df.drop(UnixTime, axis=1, inplace=True)
        if config.current_mode_is_test:
            ...
            # self.save_excel("testpytest2.xlsx")
        return df
    def save_excel(self, name):
        if isinstance(self.df, pd.DataFrame):
            self.df.to_excel(name)
    @staticmethod
    def list_to_df(items):
        return pd.DataFrame(items)
    def make_df_float(self):
        def check_cols(x: str):
            items = (
                "Tarih" not in x,
                "WEEK" not in x,
                "DAY" not in x,
            )
            return all(items)
        columns = (x for x in self.df.columns if check_cols(x))
        other_cols = (x for x in self.df.columns if not check_cols(x))
        old_df = self.df
        try:
            self.df = self.df[columns].astype(float)
        except:
            # print_with_failure_style("Could not convert some columns to float type...")
            ...
        for other_column in other_cols:
            self.df[other_column] = old_df[other_column]
    def convert_to_df_abstract(self):
        from rich import inspect
        inspect(self.data_model)
        obj = {
            "csv": self.convert_csv_df,
            "json": self.convert_json__df
        }
        fn = obj[self.data_model.type_name]
        fn()
    def convert_json__df(self, json_buffer: str = None):
        # print("To be implemented soon...")
        print(json_buffer)
        exit()
        # raise NotImplementedError
    def convert_csv_df(self, csv_buffer: str = None):
        if csv_buffer is None:
            csv_buffer = self.data_model.data
        pd_rows = list(
            map(self.extract_values_from_line,
                (item for item in self.line_gen(csv_buffer))
                )
        )
        if not pd_rows:
            # print("convert_csv_df get empty buffer ")
            return
        pd_rows.pop(0)
        self.df_no_columns = self.list_to_df(pd_rows)
        try:
            self.df = self.get_columns(self.df_no_columns)
        except DfColumnsDoesNotMatch:
            print_with_failure_style("columns did not match returning df with no columns...")
            self.df = self.df_no_columns
            print(self.df.head())
            # filename = "error_" + str(id(self.df)) + ".xlsx"
            # print_with_updating_style("file was saved", filename)
            # self.df.to_excel(filename)
            print_with_info_style("This error will be checked again...")
            time.sleep(1)
            return False
        except Exception as e:
            print_with_failure_style(e)
            pass
        self.make_df_float()
        return self.df


from evdspy import get_series, get_series_exp


def test_get_proxies_env_helper(capsys):
    index = """
    TP.DK.USD.A
    TP.DK.EUR.A
    TP.DK.CHF.A
    TP.DK.GBP.A
    TP.DK.JPY.A


    """

    with capsys.disabled():
        df = get_series(
            index, cache=False, start_date="02-01-2017", end_date="31-12-2017"
        )
        print(df.head())
        res = get_series_exp(
            index, cache=False, start_date="02-01-2017", end_date="31-12-2017"
        )

        # assert df == res.data
        print(res.data)
        assert res.data.shape == df.shape
        # assert res.data.values == df.values 




from evdspy.EVDSlocal.session.session_ import *
from evdspy.EVDSlocal.log_classes.log_template import *
session = get_session()
def test_get_session(capsys):
    s = get_session()

def test_log():
    deb, deb2, debug = get_debugger()


# from evdspy.EVDSlocal.index_requests.user_requests import UrlSeries, UrlDataGroup
from evdspy.EVDSlocal.index_requests.user_requests.Request_config import RequestConfig


# ....................................................................... UrlBuilder
class UrlBuilder:
    def __init__(self,
                 config: RequestConfig,
                 url_type=None) -> None:
        self.config = config
        self.series_part = self.config.create_series_part()
        if not url_type:
            self.get_url_type()
        self.alias = self.url_type.alias

    def get_url_type(self):
        from evdspy.EVDSlocal.index_requests.user_requests import UrlSeries, UrlDataGroup
        url_type = UrlSeries()
        if self.config.check_type() == "datagroup":
            url_type = UrlDataGroup()
        self.url_type = url_type

    def create_url_for_series(cls) -> str:
        domain = cls.domain
        return f"{domain}/{cls.alias}{cls.series_part}&startDate={cls.config.start_date}&endDate={cls.config.end_date}&type=json"

    @property
    def domain(self) -> str:
        return self.url_type.domain

    @property
    def basic_url(self) -> str:
        config = self.config
        return f"{self.domain}/{self.alias}{self.series_part}&startDate={config.start_date}&endDate={config.end_date}&type=json"

    @property
    def url(self) -> str:
        config = self.config
        if config.frequency is None and config.aggregation is None and config.formulas is None:
            return self.basic_url
        """ config parts """
        formulas_str = config.formulas_to_str()
        aggregation_type_str = config.aggregation_type_to_str()
        freq_string = config.freq_str()
        """..."""
        parts = (
            f"{self.domain}/{self.alias}{self.series_part}{freq_string}{formulas_str}{aggregation_type_str}",
            f"startDate={config.start_date}",
            f"endDate={config.end_date}",
            "type=json"
        )
        return "&".join(parts)


# Implementation of HelpUrlBuilder
class HelpUrlBuilder:

    def __init__(self, config: RequestConfig, url_type=None) -> None:
        self.config = config
        self.index = self.clean(self.config.initial_index)
        # self.series_part = self.config.create_series_part()
        if not url_type:
            self.get_url_type()
        self.alias = self.url_type.alias

    def __str__(self):
        return f"""
    ........................... HelpUrlBuilder ..............
    basic_url : {self.url}
    meta_url : {self.meta_url}
    ........................... HelpUrlBuilder ..............
    """

    def get_url_type(self):
        from evdspy.EVDSlocal.index_requests.user_requests import UrlSeries
        url_type = UrlSeries()
        self.url_type = url_type

    @property
    def domain(self) -> str:
        return self.url_type.domain

    def clean(self, index: str) -> str:
        if "bie_" in index:
            return index
        return index.replace('_', '.')

    @property
    def basic_url(self) -> str:
        """
        https://evds2.tcmb.gov.tr/service/evds/serieList/type=xml&code=TP.DK.USD.A
        https://evds2.tcmb.gov.tr/service/evds/serieList/type=csv&code=bie_yssk
        https://evds2.tcmb.gov.tr/service/evds/serieList/type=json&code=TP.GSYIH02.GY.CF
        :return:
        """
        return f"https://evds2.tcmb.gov.tr/service/evds/serieList/type=json&code={self.index}"

    @property
    def url(self) -> str:
        return self.basic_url




# from .menu import yedekle_this, yedekle_this_onayisteme
import sys
# @dataclass
class GithubActions:
    def is_testing(self):
        return "hostedtoolcache" in sys.argv[0]
class PytestTesting:
    def is_testing(self):
        # print(" sys.argv[0]" ,  sys.argv[0])
        return "pytest" in sys.argv[0]
def get_input(msg, default=None):
    if GithubActions().is_testing() or PytestTesting().is_testing():
        if not default:
            print("currently testing with no default ")
            return False
        return default
    return  input(msg)


import os
import traceback

# from dotenv import load_dotenv
from pathlib import Path

try:

    from dotenv import load_dotenv


    def load_env(verbose=False):
        """Load environment variables from a .env file in the current directory."""
        dotenv_path = Path(".") / '.env'
        if verbose:
            print(f"reading apikey from [{dotenv_path}]")
        load_dotenv(dotenv_path)

except Exception:

    print("""[solution]
          $ pip install python-dotenv 
          """)
    traceback.print_exc()

    raise ModuleNotFoundError()




from abc import ABC
from dataclasses import dataclass
@dataclass
class DataModel(ABC):
    data: any
    type_name: str = ""
@dataclass
class DataModelCSV(DataModel):
    type_name: str = "csv"
@dataclass
class DataModelJSON(DataModel):
    type_name: str = "json"



# ------------------------------------------------------------------------------
from ..common.files import *
from dataclasses import dataclass
from .series_examples import *
from ..components.evds_seri_files import EvdsSeriesRequestWrapper, EvdsSeriesRequest, EvdsSeri
from evdspy.EVDSlocal.series_format.populate_series import *
from evdspy.EVDSlocal.config.config import default_series_file_name
import random
from ..common.colors import *
from typing   import List 

GSEP = "--"
# ------------------------------------------v1.0 refactored------------------------------------------------------
from pathlib import Path
def get_approval(file_name):
    msg = f"{file_name} already exists. \n If you would like to replace it with a new example please type `replace`: "
    print_get_input_style(msg )
    ans = input()
    if str(ans).lower().strip() == "replace":
        return True
    return False
def create_series_text_file(file_name: str, content=""):
    """General usage"""
    if Path(file_name).is_file() and "-locked" not in file_name.lower():
        if not get_approval(file_name):
            return False
    print_with_creating_style(f"creating...{file_name}")
    result, msg = Write(file_name, content)
    print_with_success_style(msg)
    time.sleep(1)
    return result
def create_series_text_example(file_name: str = default_series_file_name, onsetup: bool = False):
    if onsetup:
        if file_exists_show_and_return(file_name):
            return
    """ main series config file """
    return create_series_text_file(file_name, random.choice(example_series))
# ------------------------------------------ / v1.0 refactored------------------------------------------------------
def series_file_info(file_name):
    print_with_info_style(f"In order to use your own choice of data series in "
                          f"`{file_name}` you may change this file accordingly")
from evdspy.EVDSlocal.components.bucket_from_series import BucketFromSeriesFile
def get_locked_file_title():
    title = """
#package : evdspy
#author  : sermet.pekin
#Series_config_file_LOCKED :Please do not modify.
#
#						E V D S P Y  _  C O N F I G  _  F I L E (LOCKED) ---------------------------------------------
#
# 	-----------------------------------------------------------------------------------------
#		NOTE : PLEASE DO NOT MODIFY *LOCKED* VERSION OF THE CONFIG FILE (THIS FILE).
#       BECAUSE IT WILL NOT HAVE EFFECT ON RUN.
#       INSTEAD IF YOU WOULD LIKE TO MODIFY ANY PROPERTIES OF SERIES OR OTHER OPTIONS
#       YOU MAY MODIFY UNLOCKED VERSION OF THIS FILE. (`config_series.cfg`)
#
#       THIS FILE(LOCKED) IS HELP USERS TO SEE HOW MACHINE MODIFIED DEFAULT OPTIONS
#       DURING RUN AFTER CHECKING PROCESS.
#
# -------------------------------------------------------------------------------------------
"""
    return title
from evdspy.EVDSlocal.setup_project.user_series_config import frequency_dict, formulas_dict
from evdspy.EVDSlocal.state.current_state import CurrentState
from typing import Dict
from evdspy.EVDSlocal.components.request_or_cache import RequestOrCacheResultInfo
def get_request_results():
    result_data_dict: Dict[str, RequestOrCacheResultInfo] = CurrentState().get_result_dict()
    items = result_data_dict.items()
    items_str_list = (tuple_request_or_cache_info[1].get_data() for tuple_request_or_cache_info in items)
    result_content = "\n".join(items_str_list)
    result = f"""
------------REQUEST or CACHE RESULTS------------------
{result_content}
------------/REQUEST or CACHE RESULTS------------------
"""
    return result
def create_content(bfs: BucketFromSeriesFile, evds_series_request_: EvdsSeriesRequest, subject,
                   series_list: List[EvdsSeri]):
    list_ = [x.ID for x in series_list]
    series_list_str = "\n".join(list_)
    r = f"""
---Series---------------------------------
foldername : {bfs.folder_name}
abs_path : {bfs.abs_path}
subject  : {subject}
prefix   : {bfs.prefix}
frequency : {bfs.frequency}                # {frequency_dict.get(int(bfs.frequency), 'Not Found (request will be made with default option for this property)')}
formulas : {bfs.formulas}                  # {formulas_dict.get(int(bfs.formulas), 'Not Found (request will be made with default option for this property)')}
aggregateType : {bfs.aggregateType}
------------SERIES CODES------------------
{series_list_str}
------------/SERIES CODES------------------
---/Series---------------------------------
--++--
"""
    return r
def create_series_file_from_Wrapper(wrapperList: List[EvdsSeriesRequestWrapper]):
    contents = []
    for EvdsSeriesRequestWrapper_ in wrapperList:
        name = EvdsSeriesRequestWrapper_.name
        subject = EvdsSeriesRequestWrapper_.subject
        series_list = EvdsSeriesRequestWrapper_.EvdsSeriesRequest_.series_list
        contents.append(
                create_content(EvdsSeriesRequestWrapper_.bfs, EvdsSeriesRequestWrapper_.EvdsSeriesRequest_, subject,
                               series_list))
    cont = get_locked_file_title() \
           + f"\n{GSEP}".join(contents) \
           + get_request_results()
    return cont
from ..components.evds_seri_files import *
# ------------------------------------------------------------------------------
# /*
#               SeriesFileFormat
#
# ------------------------------------------------------------------------------
# */
@dataclass
class SeriesFileFormat:
    series_filename: str
    EvdsSeriesRequestWrapperList: any = None
    def __post_init__(self):
        self.get_series_from_file()
    def read(self):
        return Read(self.series_filename, f"{self.series_filename} problem")
    def get_series_from_file(self):
        self.EvdsSeriesRequestWrapperList: \
            List[EvdsSeriesRequestWrapper] = \
            PopulateSeries(
                    self.series_filename
            ).split_series_file()
        return self.EvdsSeriesRequestWrapperList
def test_SeriesFileFormat2():
    sff = SeriesFileFormat("series.txt")
    return sff.EvdsSeriesRequestWrapperList  # s.read()
# ------------------------------------------------------------------------------
# /*
#               SeriesCreator
#
# ------------------------------------------------------------------------------
# */
@dataclass
class SeriesCreator:
    def create(self):
        ...



from dataclasses import dataclass
from typing import Union, Tuple
from evdspy.EVDSlocal.messages.error_classes import ApiKeyNotSetError, SeriesEmptyError
from evdspy.EVDSlocal.config.config import config
from evdspy.EVDSlocal.requests_.real_requests import RealRequestWithParam
CANCEL_REQUEST_TEMP = config.cancel_request_temp
def test():
    p = PrepareUrl(series=('TP_ODEMGZS_NORVEC-8',), frequency=5, api_key='api_key')
    print(p.get_data())
from rich import print
def basic_for_test(api_key):
    url = "https://evds2.tcmb.gov.tr/service/evds/series=TP.ODEMGZS.BDTTOPLAM-TP.ODEMGZS.ABD-TP.ODEMGZS.ARJANTIN-TP.ODEMGZS.BREZILYA-TP.ODEMGZS.KANADA-TP.ODEMGZS.KOLOMBIYA-TP.ODEMGZS.MEKSIKA-TP.ODEMGZS.SILI&startDate=01-01-2019&endDate=01-12-2030&frequency=5&aggregationTypes=avg-avg-avg-avg-avg-avg-avg-avg&formulas=0-0-0-0-0-0-0-0&type=csv"
    # url = f"{url}&key={api_key}"
    req = RealRequestWithParam(url, api_key=api_key)
    response = req.request()
    # p_basic = PrepareUrl()
    # response = p_basic.get_data_with_url(url)
    return getattr(response, "status_code") and response.status_code in (200,)
def basic_for_testOLD(api_key):
    url = "https://evds2.tcmb.gov.tr/service/evds/series=TP.ODEMGZS.BDTTOPLAM-TP.ODEMGZS.ABD-TP.ODEMGZS.ARJANTIN-TP.ODEMGZS.BREZILYA-TP.ODEMGZS.KANADA-TP.ODEMGZS.KOLOMBIYA-TP.ODEMGZS.MEKSIKA-TP.ODEMGZS.SILI&startDate=01-01-2019&endDate=01-12-2030&frequency=5&aggregationTypes=avg-avg-avg-avg-avg-avg-avg-avg&formulas=0-0-0-0-0-0-0-0&type=csv"
    url = f"{url}&key={api_key}"
    p_basic = PrepareUrl()
    response = p_basic.get_data_with_url(url)
    return getattr(response, "status_code") and response.status_code in (200,)
def basic_for_test2(api_key):
    p_basic = PrepareUrl(
        series=('TP.ODEAYRSUNUM6.Q1',),
        api_key=api_key,
        frequency=5,
        basic=True)
    print(p_basic.get_data())
    url = p_basic.get_data()
    url = clean(url)
    print(url)
    return "WHAT"
    # make_request
    return p_basic.response.status_code in (200,)
def basic_test_with_real_key():
    real_api_key = ''
    return basic_for_test(real_api_key)
def clean(url):
    import string
    return url.translate({ord(c): None for c in string.whitespace})
default_start_date = "01-01-2019"
default_end_date = "01-01-2030"
@dataclass
class PrepareUrl:
    series: Tuple[str] = ()
    api_key: str = ""
    frequency: Union[int, None] = None
    aggregateType: Union[str, None] = "avg"
    basic: bool = False
    def get_data(self,
                 series=None,
                 api_key=None,
                 startDate=default_start_date,
                 endDate=default_end_date,
                 frequency=None,
                 aggregateType=None
                 ):
        if series is None:
            series = self.series
        if api_key is None:
            api_key = self.api_key
        if frequency is None:
            frequency = self.frequency
        if self.basic:
            self.create_url(
                series=series,
                api_key=api_key,
                startDate=startDate,
                endDate=endDate,
                frequency=frequency,
                aggregateType=aggregateType,
            )
        return self.get(
            self.create_url(
                series=series,
                api_key=api_key,
                startDate=startDate,
                endDate=endDate,
                frequency=frequency,
                aggregateType=aggregateType,
            ))
    def get(self, url):
        print(url)
        return url
    def get_data_with_url(self, url):
        return self.make_request(url)
    def series_to_str(self, series):
        return "-".join(series)
    def create_url(self, series, api_key, startDate, endDate, frequency, aggregateType):
        domain: str = "https://evds2.tcmb.gov.tr/service/evds/"
        if series is None:
            raise SeriesEmptyError
        if api_key is None:
            raise ApiKeyNotSetError
        if frequency is None and aggregateType is None:
            return f"{domain}series={self.series_to_str(series)}&startDate={startDate}&endDate={endDate}&type=csv&key={api_key}"
        if frequency:
            if aggregateType:
                return f"{domain}series={self.series_to_str(series)}" \
                       f"&frequency={frequency}" \
                       f"&aggregationTypes={aggregateType}" \
                       f"&type=csv" \
                       f"&key={api_key}"
            else:
                return f"{domain}series={self.series_to_str(series)}" \
                       f"&frequency={frequency}" \
                       f"&aggregationTypes=avg" \
                       f"&type=csv" \
                       f"&key={api_key}"
    def make_request(self, url=None):
        from ..messages.error_classes import ApiKeyNotSetError
        if url is None:
            url = self.url
        if CANCEL_REQUEST_TEMP:
            print("CANCEL_REQUEST_TEMP is True ...prepare.151")
            return False
        try:
            # self.response = requests.get(url)
            self.response = RealRequestWithParam(url).request()
        except ApiKeyNotSetError:
            ...
            return
        return self.response
import requests
def make_request(url: str):
    return requests.get(url)


# ------------------------------------------------------------------------------
import requests
from evdspy.EVDSlocal.components.url_class import URLClass
from evdspy.EVDSlocal.console.proxy_for_menu import get_proxies_env
from evdspy.EVDSlocal.utils.utils_general import *
from evdspy.EVDSlocal.requests_.my_cache import MyCache, save_pickle_for_test, lru_cache_patched, load_test_pickle
from evdspy.EVDSlocal.components.options_class import Options
from evdspy.EVDSlocal.config.credentials_file import Credentials
from evdspy.EVDSlocal.initial.start_options import current_mode_is_test
from evdspy.EVDSlocal.initial.start_args import Args
from evdspy.EVDSlocal.messages.error_classes import *
from evdspy.EVDSlocal.config.apikey_class import *
from evdspy.EVDSlocal.config.config import config
from evdspy.EVDSlocal.stats.save_stats_of_requests import *
from evdspy.EVDSlocal.requests_.request_error_classes import REQUEST_ERROR_CODES
from evdspy.EVDSlocal.common.url_clean import remove_api_key
from dataclasses import dataclass
from evdspy.EVDSlocal.requests_.mock_req import *
from evdspy.EVDSlocal.requests_.real_requests import *
from typing import Optional
from dataclasses import dataclass, field

# ------------------------------------------------------------------------------
m_cache = MyCache()
CANCEL_REQUEST_TEMP = config.cancel_request_temp


def do_first_true_order(funcs, preds, url):
    conds = zip(funcs, preds)
    for func, pred in conds:
        if isinstance(pred, bool) and pred:
            v = func(url)
            return v
    return False


def decide_request_for_test_real(url: str, proxies=None):
    """ Finally ready to make request """
    if config.current_mode_is_test and not config.temp_cancel_mock_request:
        """ pytest will get this mock response object"""
        return mock_request(url, proxies)
    # print(f"WARNING : mock results are coming {url}")
    # return mock_request(url, proxies)
    if CANCEL_REQUEST_TEMP:
        print("request was cancelled...")
        return mock_request(url, proxies)
    if proxies is None:
        return requests.get(url)
    return requests.get(url, proxies=proxies)


@dataclass
class EVRequest:
    options_: Options
    session: any = None
    proxy: Optional[str] = False
    args: Args = field(default_factory=Args)  # None  # Args(tuple(sys.argv))
    last_url_checked: str = "NoneURL"
    URL_Instance: field(default_factory=URLClass) = field(default_factory=URLClass)
    force_no_cache = False

    # report : post init
    def __post_init__(self):
        # self.URL_Instance = field(default_factory=URLClass) # None
        self.credentials = Credentials(self.options_)
        """
        we are now reporting how many times
        we requested new items and how many times we avoided making new
        request by using cache results...
        """
        self.report = Report("stats_requests.txt")

    def get_request_with_proxy(self, url: str, proxy: str):
        proxies = {
            'http': proxy,
            'https': proxy,
        }
        return decide_request_for_test_real(url, proxies=proxies)

    def get_proxies(self):
        env_proxies = get_proxies_env()
        if self.args.proxy is None:
            if env_proxies:
                return env_proxies
            return None
        proxy = self.args.proxy
        proxies = {
            'http': proxy,
            'https': proxy,
        }
        return proxies

    def proxy_from_file(self, url: str):
        print("using proxy from file")
        return self.get_request_with_proxy(url, self.args.proxy)

    def proxy_from_cmd_line(self, url: str):
        print("using proxy from commandline")
        return self.get_request_with_proxy(url, self.args.proxy)

    def proxy_from_env(self, url: str):
        print("using proxy from .env file ")
        env_proxies = get_proxies_env()
        if env_proxies:
            return self.get_request_with_proxy(url, env_proxies)
        return self.proxy_from_cmd_line(url)

    def check_any_apikey(self):
        if not config.current_mode_is_test:
            return True
        if not ApikeyClass().get_valid_api_key():
            raise ApiKeyNotSetError
        return True

    def check_if_request_ok(self, status_code: int):
        corrects = (200,)
        if status_code in corrects:
            self.report.add(f"{status_code} returned. ")
            return True
        self.report.add(f"{status_code} returned. ")
        ExceptionClass = REQUEST_ERROR_CODES.get(status_code, None)
        if ExceptionClass:
            print(ExceptionClass)
            print(f"STATUS CODE : {status_code}\n url : {remove_api_key(self.last_url_checked)}")
            time.sleep(2)
            # raise ExceptionClass(self.__doc__)
        else:
            print(
                f"Program was checking request status code and noticed "
                f"{status_code} not included in  `REQUEST_ERROR_CODES`"
            )
        self.report.add(f"{status_code} returned. ")
        return False

    def post_urls(self, url: str) -> tuple:
        # url = url.translate({ord(c): None for c in string.whitespace})
        url = URL_temizle(url)  # whitespace etc.
        no_apikey_url = remove_api_key(url)  # remove key for logs
        self.report.add(f"{no_apikey_url} will be requested. ")
        self.last_url_checked = url
        return (url, no_apikey_url)

    def get_request_alternatives(self, url):
        result = do_first_true_order(
            funcs=[
                self.proxy_from_env,
                self.proxy_from_cmd_line,
                self.proxy_from_file,
                decide_request_for_test_real
            ],
            preds=[
                get_proxies_env(),
                self.args.proxy,
                self.credentials.proxy,
                True],
            url=url
        )
        return result

    def check_result(self, result):
        if any(
                (
                        not hasattr(result, "status_code"),
                        not isinstance(result.status_code, int),
                        not self.check_if_request_ok(result.status_code)
                )
        ):
            print(f"request returned an error code {result.status_code} , url : {self.no_apikey_url}")
            return False
        return True

    def get_request_before_cache(self, url: str):
        """
        decorated version of this function needs to be
        welcomed by this func to check api key etc.
        TODO
            stats for apikey and results
            warn user if there are many 500 status_code results for the certain api key
        """
        no_apikey_url = remove_api_key(url)
        # print("checking url ... ", no_apikey_url)
        if not self.check_any_apikey():
            return False
        return self.get_request(url)

    def get_request(self, url: str):
        """
        first will check cache if there is cache
        below will not run at all
        """
        # if ApikeyClass().now_testing_is_key_is_valid    :
        #     return self.get_request_by_checking_NO_CACHE(url)
        return self.get_request_by_checking_cache(url)
        # return self.get_request_helper(url )

    @lru_cache_patched
    @m_cache.cache
    def get_request_by_checking_cache(self, url: str):
        """
        """
        return self.get_request_common(url)

    def get_request_by_checking_NO_CACHE(self, url: str):
        """
        """
        return self.get_request_common(url)

    def get_request_w_param(self, url: str, proxies=None):
        return RealRequestWithParam(url, proxies).request()

    def get_request_common(self, url: str):
        """
        first will check cache if there is cache
        below will not run at all
        """
        self.url, self.no_apikey_url = self.post_urls(url)
        if current_mode_is_test and ApikeyClass().key is False:
            print("Current Mode is test not requesting new data see initial.start_options")
            return load_test_pickle()
        """ if it is running we are definitely requesting """
        safe_url = remove_api_key(self.url)
        print("requesting!!", safe_url)
        """ Exception checks will happen in function below"""
        result = self.get_request_w_param(self.url, self.get_proxies())
        # result = self.get_request_alternatives(self.url)
        # result = self.get_request_alternatives(self.no_apikey_url)
        if not self.check_result(result):
            return False
        """ request looks solid we are ready to save """
        if result and self.check_if_request_ok(result.status_code):
            save_pickle_for_test(result)
        return result




from evdspy.EVDSlocal.setup_project.user_series_config import quickNoteForParams
def add_indent(content):
    quickNoteForParamsLines = content.split("\n")
    quickNoteForParamsLines = (" " * 10 + x for x in quickNoteForParamsLines)
    content = "\n".join(quickNoteForParamsLines)
    return content
quickNoteForParams = add_indent(quickNoteForParams)
series_title = f"""#Series_config_file
E V D S P Y  _  C O N F I G  _  F I L E  ---------------------------------------------
#
# This file will be used by evdspy package (python) in order to help updating
# your series.
# Script will be adding to it content when you setup a new project.
# Deleting this file current file may require to setup a new configuration from the beginning
#		please see `README.md` file
# ----------------------------------------------------------------------------------------
#
#About alternative params
# ----------------------------------------------------------------------------------------
{quickNoteForParams}
#Begin_series
"""



from evdspy.EVDSlocal.common.common_imports import *
from .sample_series_content import sample_series
"""
    This file runs only on startup to create some samples to check setup.
    you may change `series.txt` content according to format provided both on
    readme file and in the text file.
"""
from evdspy.EVDSlocal.utils.utils_general import *
from evdspy.EVDSlocal.initial.start_options import *
def create_sample_series_text_file():
    with open(default_series_file_name, "w", encoding=ENCODING) as file_:
        file_.write(sample_series)
def on_start_create():
    from pathlib import Path
    f = Path(default_series_file_name)
    if not f.is_file():
        create_sample_series_text_file()
        print(f"{indent} sample file {default_series_file_name} was created...")
__all__ = [
    'on_start_create'
]


# ................................................................... RequestConfig
from dataclasses import dataclass
from typing import Union, Optional

# from evdspy import default_start_date_fnc, default_end_date_fnc
from evdspy.EVDSlocal.components.excel_class import replace_all
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import freq_enum, correct_types, AggregationType, \
    Formulas, default_start_date_fnc, default_end_date_fnc
from evdspy.EVDSlocal.index_requests.user_requests.Serialize import Serialize
from evdspy.EVDSlocal.index_requests.user_requests.User_req_typings import T_tuple_str_int_None, T_str_tuple_None, \
    T_str_int_None
from evdspy.EVDSlocal.index_requests.user_requests.User_request_utils import UserRequestUtils


@dataclass
class RequestConfig(Serialize):
    index: Union[str, tuple[str, ...], list[str, ...]]
    start_date: str = default_start_date_fnc()
    end_date: str = default_end_date_fnc()
    frequency: T_str_int_None = None
    formulas: T_tuple_str_int_None = None
    aggregation: T_str_tuple_None = None
    cache: bool = False
    cache_name: str = ""
    initial_index = None
    # series_url_instance: SeriesUrlClass = SeriesUrlClass()
    """
    Represents a user request for fetching data series from a specified source.
    This class encapsulates the details necessary to construct and execute a data retrieval request,
    handling various configurations like date ranges, frequency, and data aggregation types. It
    also manages proxy settings and caching mechanisms to optimize data retrieval operations.
    Attributes:
        index (Union[str, Tuple[str]]): The identifier(s) for the data series to fetch.
        start_date (str): The start date for the data retrieval in 'DD-MM-YYYY' format.
        end_date (str): The end date for the data retrieval in 'DD-MM-YYYY' format.
        frequency (Union[str, int, None]): The frequency at which data should be retrieved.
        formulas (Union[str, int, Tuple[str, int], None]): The calculation types to apply to the data.
        aggregation (Union[str, Tuple[str], None]): The aggregation methods to apply to the data.
        cache (bool): Enables or disables caching of the request results.
        cache_name (str): The name under which to store the cached results, if caching is enabled.
        series_url_instance (SeriesUrlClass): An instance of SeriesUrlClass to handle URL creation.
    Methods:
        __post_init__: Initializes further attributes and performs checks after the basic initialization.
    Raises:
        ValueError: If the index is in an invalid format or necessary attributes are missing.
    """

    def check_type(self):
        if UserRequestUtils.looks_like_datagroup(self.initial_index):
            return "datagroup"
        return "series"

    def __post_init__(self):
        self.initial_index = self.index
        self.correct_index()
        self.correct_formulas_aggr()
        self.check()

    def upper_index(self) -> None:
        def upper(string: str) -> str:
            if UserRequestUtils.looks_like_datagroup(self.initial_index):
                return string
            return string.upper()

        self.index = tuple(map(upper, self.index))

    def clean(self) -> None:
        self.index = tuple(map(UserRequestUtils.clean_chars, self.index))

    def correct_index(self):
        if "\n" in self.index or "\t" in self.index:
            self.index = self.template_to_tuple(self.index)
        self.index = tuple([self.index]) if not isinstance(self.index, (tuple, list,)) else self.index
        self.clean()
        self.upper_index()
        self.check_index()

    def correct_formulas_aggr(self):
        self.formulas = self.correct_type_to_tuple(self.formulas)
        self.aggregation = self.correct_type_to_tuple(self.aggregation)

    def correct_type_to_tuple(self, value: any) -> Optional[tuple]:
        if value is None:
            return None
        if isinstance(value, (list, tuple)):
            if len(value) == len(self.index):
                return value
        if isinstance(value, (str,)):
            return tuple(value for _ in self.index)
        return tuple(value[0] for _ in self.index)

    @staticmethod
    def template_to_tuple(index: str) -> tuple:
        def clean(string: str):
            return string.split("#")[0].strip() if len(string.split("#")) > 0 else None

        index = replace_all(index, "\t", "\n")
        index_tuple = index.splitlines()
        t = tuple(clean(x) for x in index_tuple)
        return tuple(x for x in t if x is not None and len(x) > 3)

    def check_index(self) -> None:
        if isinstance(self.index, (int, float,)):
            raise ValueError("index must be a string ot tuple of string ")

    def freq_str(self) -> str:
        if self.frequency is None:
            return ""
        if isinstance(self.frequency, int):
            return f"&frequency={self.frequency}"
        return freq_enum(self.frequency)

    def agr_form_type_to_str(self, value: Optional[tuple], part_name="aggregationTypes"):
        if value is None:
            return ""
        value = tuple(map(str, value))
        string = "-".join(value)
        return f"&{part_name}=" + string

    def aggregation_type_to_str(self) -> str:
        self.aggregation = correct_types(self.aggregation, enum_class=AggregationType)
        return self.agr_form_type_to_str(self.aggregation, "aggregationTypes")

    def formulas_to_str(self) -> str:
        self.formulas = correct_types(self.formulas, enum_class=Formulas)
        return self.agr_form_type_to_str(self.formulas, "formulas")

    def check(self) -> None:
        if self.formulas is not None:
            assert len(self.formulas) == len(self.index)
        if self.aggregation is not None:
            assert len(self.aggregation) == len(self.index)
        if self.frequency is not None:
            assert isinstance(self.frequency, (int, str,))

    def create_series_part(self) -> str:
        indexes = self.index
        if isinstance(indexes, str):
            indexes = tuple([indexes])
        return "-".join(indexes)



"""
TODO async menu
"""
if False:

    import asyncio
    from IPython.display import display, clear_output
    from pathlib import Path
    from .menu import check


    # Placeholder functions for the menu options
    async def check():
        pass


    async def setup():
        pass


    async def create_options_file():
        pass


    async def create_series_file():
        pass


    async def setup_series_steps():
        pass


    async def get():
        pass


    async def get_categories_main():
        pass


    async def help_():
        pass


    async def show_apikey():
        pass


    async def set_apikey_input():
        pass


    async def remove_cache():
        pass


    async def console_main_from_the_menu():
        pass


    async def version():
        pass


    async def py_version():
        pass


    async def check_compat():
        pass


    async def main_exit_function():
        pass


    # Menu item class
    class MenuItem:
        def __init__(self, function, description):
            self.function = function
            self.description = description


    # Menu maker class
    class MenuMaker:
        def __init__(self, menu_items, exit_item, exit_menu_call_back):
            self.menu_items = menu_items
            self.exit_item = exit_item
            self.exit_menu_call_back = exit_menu_call_back

        async def display(self):
            while True:
                clear_output(wait=True)
                print("Please choose an option:")
                for idx, item in enumerate(self.menu_items):
                    print(f"{idx + 1}. {item.description}")
                if self.exit_item:
                    print(f"{len(self.menu_items) + 1}. Exit")

                user_input = await asyncio.get_event_loop().run_in_executor(None, input, "Enter choice: ")
                try:
                    choice = int(user_input) - 1
                    if choice == len(self.menu_items):
                        await self.exit_menu_call_back()
                        break
                    elif 0 <= choice < len(self.menu_items):
                        await self.menu_items[choice].function()
                    else:
                        print("Invalid choice, please try again.")
                except ValueError:
                    print("Invalid input, please enter a number.")

                await asyncio.sleep(0.1)


    # Function to display the menu
    async def menu_display():
        funcs = [
            ("check setup", check),
            ("setup", setup),
            ("create user options file", create_options_file),
            ("create series file", create_series_file),
            ("add new series group", setup_series_steps),
            ("get data", get),
            ("get categories (get all series of a datagroup)", get_categories_main),  # next version
            # ("get data groups (on development)", get_datagroups_data_main),  # next version
            ("help", help_),
            ("show api key", show_apikey),
            ("save api key to file", set_apikey_input),
            ("remove cache folders", remove_cache),
            ("evdspy as a command line prompt", console_main_from_the_menu),
            ("version", version),
            ("py version", py_version),
            ("check compatibility of your python version", check_compat),
        ]

        menu_items = list(map(lambda x: MenuItem(x[1], x[0]), funcs))
        menu_maker = MenuMaker(
            menu_items=menu_items,
            exit_item=True,
            exit_menu_call_back=main_exit_function
        )
        await menu_maker.display()


    # Helper function to start the menu
    async def menu_helper_async():
        global menu_already_displayed
        menu_already_displayed = True
        await menu_display()


    # Global variable to store the menu task
    menu_task = None


    # Function to start the menu
    def start_menu():
        global menu_task
        if menu_task and not menu_task.done():
            menu_task.cancel()
        menu_task = asyncio.create_task(menu_helper_async())


    # Function to stop the menu
    def stop_menu():
        global menu_task
        if menu_task and not menu_task.done():
            menu_task.cancel()
            menu_task = None
            print("Menu stopped.")


    # Start the menu using asyncio.run in a script
    if __name__ == "__main__":
        asyncio.run(menu_helper_async())

    # In Jupyter Notebook, call start_menu and stop_menu functions
    # start_menu()
    # stop_menu()




from evdspy.EVDSlocal.requests_.ev_request import *
from ..tests.core_options_test import *


def test_mock_request2(capsys):
    response = mock_request("test_url")
    with capsys.disabled():
        if verbose:
            #print(response , type(response))
            print(type(response ))
    assert isinstance(response, requests.models.Response)
def test_report2():
    report_test()



from ..components.url_class import *
from ..components.api_params import *
from ..components.api_params import EVDSApiDomainCategories
from ..config.apikey_class import ApikeyClass
from dataclasses import dataclass
from ..requests_.ev_request import EVRequest
from ..messages.error_classes import ApiKeyNotSetError
import typing as t
@dataclass
class GeneralIndexesBase(ABC):
    """GeneralIndexesBase"""
    domain: t.Union[str, EVDSApiDomain]
    EVRequest_: EVRequest = None
    type_param: dataTypeEnum = dataTypeEnum.csv  # csv or json
    mode: ModeEnumDatagroup = ModeEnumDatagroup.all_groups
    code: int = 1
    # EVRequest_: EVRequest = EVRequest()
    def __repr__(self):
        res = f"""
class : [{self.__class__.__name__} ]
mode :  [{self.mode}]
code : [{self.code}]
url : [{self.EVRequest_.URL_Instance.url}]
"""
        return res
    def __str__(self):
        res = f"""
class : [{self.__class__.__name__} ]
mode :  [{self.mode}]
code : [{self.code}]
url : [{self.EVRequest_.URL_Instance.url}]
"""
        return res
    def create_url_first_part(self):
        self.api_key = ApikeyClass().get_valid_api_key()
        full_list = []
        self.complete_url_instance = URLClass(full_list, domain=str(self.domain))
        self.complete_url_instance.add_apikey(self.api_key)
        # self.complete_url_instance.add_item(dataTypeParam(self.type_param))
    def create_url(self):
        self.create_url_first_part()
        self.add_extra_params()
        self.complete_url_instance.refresh_url()
        self.complete_url_instance.create_report()
        self.EVRequest_.URL_Instance = self.complete_url_instance
        if not self.api_key:
            raise ApiKeyNotSetError
        return self.complete_url_instance.url
    def add_extra_params(self, mode: int = 1, code: t.Union[int, str] = None):
        ...
    def get_csv(self, url: str = None):
        self.create_url()
        if url is None:
            url = self.EVRequest_.URL_Instance.url
        self.EVRequest_.URL_Instance = self.complete_url_instance
        result = self.EVRequest_.get_request_before_cache(f"{url}&type=csv")
        if hasattr(result, "text"):
            self.csv = result.text
            return result.text
        return False
    def get_json(self, url: str = None):
        # raise NotImplementedError
        self.create_url()
        if url is None:
            url = self.EVRequest_.URL_Instance.url
        self.EVRequest_.URL_Instance = self.complete_url_instance
        result = self.EVRequest_.get_request_before_cache(f"{url}&type=json")
        if hasattr(result, "json"):
            self.csv = result.json()
            return result.json()
        return False
class dataGroupsNeedsCode(BaseException):
    """dataGroupsNeedsCode"""
@dataclass
class GeneralIndexesCategories(GeneralIndexesBase):
    """GeneralIndexesCategories"""
    domain: t.Union[str, EVDSApiDomain] = field(default_factory=EVDSApiDomainCategories)
@dataclass
class GeneralIndexesDatagroups(GeneralIndexesBase):
    """GeneralIndexesDatagroups"""
    domain: t.Union[str, EVDSApiDomain] = field(default_factory=EVDSApiDomainDatagroups)
    def add_extra_params(self, mode: int = 2, code: int = None):
        self.complete_url_instance.add_mode(mode)
        if code is None:
            code = self.code
        self.complete_url_instance.add_code(code)
@dataclass
class GeneralIndexesDatagroupIndividual(GeneralIndexesBase):
    """GeneralIndexesDatagroups"""
    domain: t.Union[str, EVDSApiDomain] = field(default_factory=EVDSApiDomainDatagroupIndividual)
    def add_extra_params(self, mode: int = 2, code: str = None):
        ...
        self.complete_url_instance.add_frame(code)
        # self.complete_url_instance.add_mode(mode)
        # if code is None:
        #     code = self.code
        # self.complete_url_instance.add_code(code)
    def get_json(self, url: str = None):
        if url is None:
            url = self.complete_url_instance.url
        result = self.EVRequest_.get_request_before_cache(f"{url}&type=json")
        if hasattr(result, "json"):
            self.csv = result.json()
            return result.json()
        return False
@dataclass
class GeneralIndexesDatagroupSeriesList(GeneralIndexesDatagroupIndividual):
    """GeneralIndexesDatagroups"""
    domain: t.Union[str, EVDSApiDomain] = field(default_factory=EVDSApiDominDatagroupSeriesList)
    def add_extra_params(self, mode: int = 2, code: str = None):
        self.complete_url_instance.add_code(code)
    def get_json(self, url: str = None):
        if url is None:
            url = self.complete_url_instance.url
        result = self.EVRequest_.get_request_before_cache(f"{url}&type=json")
        if hasattr(result, "json"):
            self.csv = result.json()
            return result.json()
        return False



def test1_real_reqs():
    from evdspy.EVDSlocal.index_requests.get_series_indexes import UserRequest, default_start_date_fnc, \
        default_end_date_fnc
    """
    TP.BFTUKKRE.L001	TP.BFTUKKRE.L010	TP.BFTUKKRE.L011	TP.BFTUKKRE.L012
    """
    credits_template = ("TP.BFTUKKRE.L001", "TP.BFTUKKRE.L010", "TP.BFTUKKRE.L011", "TP.BFTUKKRE.L012",)
    ur = UserRequest("TP.ODEMGZS.BDTTOPLAM")
    print(ur.url)
    assert ur.url == "https://evds2.tcmb.gov.tr/service/evds/series=TP.ODEMGZS.BDTTOPLAM&startDate=01-01-2000&endDate=01-01-2100&type=json"
    print(ur)
    balance_of_pay = "TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD"
    cache = False
    ur = UserRequest(balance_of_pay,
                     frequency="weekly",
                     start_date=default_start_date_fnc(),
                     end_date=default_end_date_fnc(),
                     aggregation=("avg",),
                     proxy="http://127.0.0.1:8000",
                     proxies={"http": "http://127.0.0.1:8000"},
                     cache=cache)
    ur.dry_request()
    d1 = ur()
    print(d1)
    ur = UserRequest(credits_template,
                     frequency="weekly",
                     start_date=default_start_date_fnc(),
                     end_date=default_end_date_fnc(),
                     aggregation=("avg",),
                     proxy="http://127.0.0.1:8000",
                     proxies={"http": "http://127.0.0.1:8000"},
                     cache=cache)
    print(ur.url)
    # result = ur.request()
    ur.dry_request()
    d2 = ur()
    print(d2)
    return d1, d2


# ....................................................................... ProxyManager
from dataclasses import dataclass
from typing import Optional, Any
from evdspy.EVDSlocal.console.proxy_for_menu import get_proxies_env


@dataclass
class ProxyManager:
    proxy: Optional[str] = None
    proxies: Optional[dict[Any, Any]] = None

    def get_proxies(self) -> Optional[dict[Any, Any]]:
        env_proxy = get_proxies_env()
        if self.proxies is None:
            if self.proxy is None:
                if env_proxy:
                    proxies = env_proxy
                else:
                    proxies = None
            else:
                proxies = self.get_proxies_helper()
        else:
            proxies = self.proxies
        # print("proxies", proxies)
        return proxies

    def get_proxies_helper(self) -> Optional[dict[Any, Any]]:
        if self.proxy is None:
            return None
        proxy = self.proxy
        proxies = {
            'http': proxy,
            'https': proxy,
        }
        return proxies
